from ._base import App
from ._bonsai import AindBehaviorServicesBonsaiApp, BonsaiApp, BonsaiAppSettings
from ._curriculum import CurriculumApp, CurriculumSettings, CurriculumSuggestion
from ._python_script import PythonScriptApp

__all__ = [
    "App",
    "BonsaiApp",
    "AindBehaviorServicesBonsaiApp",
    "PythonScriptApp",
    "BonsaiAppSettings",
    "CurriculumApp",
    "CurriculumSettings",
    "CurriculumSuggestion",
]



import contextlib
import importlib.util
import io
import logging
import os
import sys
from pathlib import Path
from types import ModuleType

import git

REPO_ROOT = Path(__file__).parents[1]

TESTS_ASSETS = REPO_ROOT / "tests" / "assets"

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())
logging.disable(logging.CRITICAL)
os.environ["WATCHDOG_EXE"] = "watchdog.exe"
os.environ["WATCHDOG_CONFIG"] = "watchdog_config.yml"


def build_example(script_path: str) -> ModuleType:
    module_name = Path(script_path).stem
    spec = importlib.util.spec_from_file_location(module_name, script_path)
    if spec is None:
        raise ImportError(f"Can't find {script_path}")
    module = importlib.util.module_from_spec(spec)
    if spec.loader is None:
        raise ImportError(f"Can't load {script_path}")
    spec.loader.exec_module(module)
    return module


@contextlib.contextmanager
def suppress_stdout():
    original_stdout = sys.stdout
    sys.stdout = io.StringIO()
    try:
        yield
    finally:
        sys.stdout = original_stdout


class SubmoduleManager:
    _instance = None
    initialized = False

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(SubmoduleManager, cls).__new__(cls)
            cls.initialized = False
        return cls._instance

    @classmethod
    def initialize_submodules(cls, force: bool = False) -> None:
        if force or not cls.initialized:
            cls._initialize_submodules()
            cls.initialized = True

    @staticmethod
    def _initialize_submodules() -> None:
        root_repo = git.Repo(REPO_ROOT)
        root_repo.git.submodule("update", "--init", "--recursive")



from pathlib import Path
from unittest.mock import Mock, patch

import pytest
from aind_behavior_services import AindBehaviorRigModel, AindBehaviorSessionModel, AindBehaviorTaskLogicModel

from clabe import ui
from clabe.launcher import Launcher
from clabe.launcher._cli import LauncherCliArgs


class MockUiHelper(ui.UiHelper):
    def __init__(self):
        self._print_func = lambda x: None
        self._input_func = lambda x: "1"
        self._prompt_pick_from_list = Mock(return_value="")
        self._prompt_yes_no_question = Mock(return_value=True)
        self._prompt_text = Mock(return_value="")
        self._prompt_float = Mock(return_value=0.0)

    def print(self, message: str) -> None:
        return self._print_func(message)

    def input(self, prompt: str) -> str:
        return self._input_func(prompt)

    def prompt_pick_from_list(self, *args, **kwargs):
        return self._prompt_pick_from_list(*args, **kwargs)

    def prompt_yes_no_question(self, prompt: str) -> bool:
        return self._prompt_yes_no_question(prompt)

    def prompt_text(self, prompt: str) -> str:
        return self._prompt_text(prompt)

    def prompt_float(self, prompt):
        return self._prompt_float(prompt)


@pytest.fixture
def mock_ui_helper():
    return MockUiHelper()


@pytest.fixture
def mock_session():
    return AindBehaviorSessionModel(
        experiment="mock", subject="mock_subject", experiment_version="0.0.0", root_path="mock_path"
    )


@pytest.fixture
def mock_rig():
    return AindBehaviorRigModel(rig_name="mock_rig", version="0.0.0")


@pytest.fixture
def mock_task_logic():
    return AindBehaviorTaskLogicModel(version="0.0.0", task_parameters={}, name="mock_task_logic")


@pytest.fixture
def mock_base_launcher(mock_rig, mock_session, mock_task_logic, mock_ui_helper, tmp_path: Path):
    launcher_args = LauncherCliArgs(
        data_dir=tmp_path / "data",
        temp_dir=tmp_path / "temp",
        create_directories=True,
    )
    # Ensure directories exist for os.chdir
    launcher_args.data_dir.mkdir(parents=True, exist_ok=True)
    launcher_args.temp_dir.mkdir(parents=True, exist_ok=True)

    with (
        patch("clabe.launcher._base.GitRepository") as mock_git,
        patch("os.chdir"),
        patch("pathlib.Path.mkdir"),
        patch("clabe.logging_helper.add_file_handler"),
        patch("clabe.launcher.Launcher._create_directory_structure"),
        patch("clabe.launcher.Launcher.validate", return_value=True),
        patch("os.environ", {"COMPUTERNAME": "TEST_COMPUTER"}),
    ):
        mock_git.return_value.working_dir = launcher_args.data_dir
        launcher = Launcher(
            rig=mock_rig,
            session=mock_session,
            task_logic=mock_task_logic,
            ui_helper=mock_ui_helper,
            settings=launcher_args,
        )
        return launcher



import glob
import logging
import os
import subprocess
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, ClassVar, Dict, Optional, Self

import pydantic
from aind_behavior_services.utils import run_bonsai_process
from typing_extensions import override

from ..services import ServiceSettings
from ..ui import DefaultUIHelper, UiHelper
from ._base import App

logger = logging.getLogger(__name__)

VISUALIZERS_DIR = "VisualizerLayouts"

if TYPE_CHECKING:
    from ..launcher import Launcher
else:
    Launcher = Any


class BonsaiAppSettings(ServiceSettings):
    """
    Settings for the BonsaiApp.

    Attributes:
        workflow (os.PathLike): Path to the Bonsai workflow file.
        executable (os.PathLike): Path to the Bonsai executable.
        is_editor_mode (bool): Whether to run Bonsai in editor mode.
        is_start_flag (bool): Whether to use the start flag when running Bonsai.
        layout (Optional[os.PathLike]): Path to the visualizer layout file.
        layout_dir (Optional[os.PathLike]): Directory containing visualizer layouts.
        additional_properties (Optional[Dict[str, str]]): Additional properties to pass to Bonsai.
        cwd (Optional[os.PathLike]): Working directory for the Bonsai process.
        timeout (Optional[float]): Timeout for the Bonsai process.
        print_cmd (bool): Whether to print the command being executed.
    """

    __yml_section__: ClassVar[Optional[str]] = "bonsai_app"

    workflow: os.PathLike
    executable: os.PathLike = Path("./bonsai/bonsai.exe")
    is_editor_mode: bool = True
    is_start_flag: bool = True
    layout: Optional[os.PathLike] = None
    layout_dir: Optional[os.PathLike] = None
    additional_properties: Dict[str, str] = pydantic.Field(default_factory=dict)
    cwd: Optional[os.PathLike] = None
    timeout: Optional[float] = None
    print_cmd: bool = False

    @pydantic.field_validator("workflow", "executable", mode="after", check_fields=True)
    @classmethod
    def _resolve_path(cls, value: os.PathLike) -> os.PathLike:
        """Resolves the path to an absolute path."""
        return Path(value).resolve()

    @pydantic.model_validator(mode="after")
    def _set_start_flag(self) -> Self:
        """Ensures that the start flag is set correctly based on the editor mode"""
        self.is_start_flag = self.is_start_flag if not self.is_editor_mode else True
        return self


class BonsaiApp(App):
    """
    A class to manage the execution of Bonsai workflows.

    Attributes:
        settings (BonsaiAppSettings): Settings for the Bonsai App
        ui_helper (UiHelper): Helper for user interface interactions.
        _result (Optional[subprocess.CompletedProcess]): Result of the Bonsai process execution.
    """

    def __init__(
        self,
        /,
        settings: BonsaiAppSettings,
        ui_helper: Optional[UiHelper] = None,
        **kwargs,
    ) -> None:
        """
        Initializes the BonsaiApp instance.

        Args:
            settings (BonsaiAppSettings): Settings for the Bonsai App.
            ui_helper (Optional[UiHelper]): UI helper instance. Defaults to DefaultUIHelper.
            **kwargs: Additional keyword arguments.

        Example:
            ```python
            # Create and run a Bonsai app
            app = BonsaiApp(settings=BonsaiAppSettings(workflow="workflow.bonsai"))
            app.run()
            # Create with custom settings
            app = BonsaiApp(
                settings=BonsaiAppSettings(
                    workflow="workflow.bonsai",
                    is_editor_mode=False,
                )
            )
            ```
        """
        self.settings = settings
        self._result: Optional[subprocess.CompletedProcess] = None
        self.ui_helper = ui_helper if ui_helper is not None else DefaultUIHelper()

    @property
    def result(self) -> subprocess.CompletedProcess:
        """
        Returns the result of the Bonsai process execution.

        Returns:
            subprocess.CompletedProcess: The result of the Bonsai process.

        Raises:
            RuntimeError: If the app has not been run yet.
        """
        if self._result is None:
            raise RuntimeError("The app has not been run yet.")
        return self._result

    @override
    def add_app_settings(self, **kwargs):
        """
        Adds application-specific settings to the additional properties.

        Args:
            **kwargs: Additional keyword arguments.

        Returns:
            Self: The updated instance of BonsaiApp.
        """

        if self.settings.additional_properties is not None:
            self.settings.additional_properties.update(**kwargs)
        else:
            self.settings.additional_properties = kwargs
        return self

    def validate(self, *args, **kwargs) -> bool:
        """
        Validates the existence of required files and directories.

        Args:
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.

        Returns:
            bool: True if validation is successful.

        Raises:
            FileNotFoundError: If any required file or directory is missing.
        """
        if not Path(self.settings.executable).exists():
            raise FileNotFoundError(f"Executable not found: {self.settings.executable}")
        if not Path(self.settings.workflow).exists():
            raise FileNotFoundError(f"Workflow file not found: {self.settings.workflow}")
        if self.settings.layout and not Path(self.settings.layout).exists():
            raise FileNotFoundError(f"Layout file not found: {self.settings.layout}")
        if self.settings.layout_dir and not Path(self.settings.layout_dir).exists():
            raise FileNotFoundError(f"Layout directory not found: {self.settings.layout_dir}")
        return True

    @override
    def run(self) -> subprocess.CompletedProcess:
        """
        Runs the Bonsai process.

        Returns:
            subprocess.CompletedProcess: The result of the Bonsai process execution.

        Raises:
            FileNotFoundError: If validation fails.
        """
        self.validate()
        self.prompt_input()

        if self.settings.is_editor_mode:
            logger.warning("Bonsai is running in editor mode. Cannot assert successful completion.")
        logger.info("Bonsai process running...")
        proc = run_bonsai_process(
            workflow_file=self.settings.workflow,
            bonsai_exe=self.settings.executable,
            is_editor_mode=self.settings.is_editor_mode,
            is_start_flag=self.settings.is_start_flag,
            layout=self.settings.layout,
            additional_properties=self.settings.additional_properties,
            cwd=self.settings.cwd,
            timeout=self.settings.timeout,
            print_cmd=self.settings.print_cmd,
        )
        self._result = proc
        logger.info("Bonsai process completed.")
        return proc

    @override
    def output_from_result(self, *, allow_stderr: Optional[bool]) -> Self:
        """
        Processes the output from the Bonsai process result.

        Args:
            allow_stderr (Optional[bool]): Whether to allow stderr output.

        Returns:
            Self: The updated instance of BonsaiApp.

        Raises:
            subprocess.CalledProcessError: If the process exits with an error.
        """
        proc = self.result
        try:
            proc.check_returncode()
        except subprocess.CalledProcessError as e:
            self._log_process_std_output("Bonsai", proc)
            raise e
        else:
            logger.info("Result from bonsai process is valid.")
            self._log_process_std_output("Bonsai", proc)

            if len(proc.stdout) > 0:
                logger.error("Bonsai process finished with errors.")
                if allow_stderr is None:
                    allow_stderr = self.ui_helper.prompt_yes_no_question("Would you like to see the error message?")
                if allow_stderr is False:
                    raise subprocess.CalledProcessError(1, proc.args)
        return self

    def prompt_visualizer_layout_input(
        self,
        directory: Optional[os.PathLike] = None,
    ) -> Optional[str | os.PathLike]:
        """
        Prompts the user to select a visualizer layout.

        Args:
            directory (Optional[os.PathLike]): Directory containing visualizer layouts.

        Returns:
            Optional[str | os.PathLike]: The selected layout file path, or None if no layout is selected.
        """
        if directory is None:
            directory = self.settings.layout_dir
        else:
            directory = Path(os.path.join(directory, VISUALIZERS_DIR, os.environ["COMPUTERNAME"]))

        layout_schemas_path = directory if directory is not None else self.settings.layout_dir
        available_layouts = glob.glob(os.path.join(str(layout_schemas_path), "*.bonsai.layout"))

        if len(available_layouts) == 0:
            logger.warning("No visualizer layouts found.")
            return None

        picked: Optional[str | os.PathLike] = None
        has_pick = False
        while has_pick is False:
            try:
                picked = self.ui_helper.prompt_pick_from_list(
                    value=available_layouts, prompt="Pick a visualizer layout:"
                )
                picked = picked if picked else ""
                has_pick = True
            except ValueError as e:
                logger.info("Invalid choice. Try again. %s", e)

        return Path(picked) if picked else None

    def _log_process_std_output(self, process_name: str, proc: subprocess.CompletedProcess) -> None:
        """
        Logs the standard output and error of a process.

        Args:
            process_name (str): Name of the process.
            proc (subprocess.CompletedProcess): The process result.
        """
        if len(proc.stdout) > 0:
            logger.info("%s full stdout dump: \n%s", process_name, proc.stdout)
        if len(proc.stderr) > 0:
            logger.error("%s full stderr dump: \n%s", process_name, proc.stderr)

    def prompt_input(self, *args, **kwargs):
        """
        Prompts the user for input if required.

        Args:
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.

        Returns:
            Self: The updated instance of BonsaiApp.
        """
        layout_dir = kwargs.pop("layout_directory", None)
        if self.settings.layout is None:
            r = self.prompt_visualizer_layout_input(layout_dir if layout_dir else self.settings.layout_dir)
            self.settings.layout = Path(r) if r else None
        return self

    def build_runner(self, allow_std_error: bool = False) -> Callable[[Launcher], Self]:
        """
        Builds a runner function for the application.

        This method returns a callable that can be executed by the launcher to run the application.

        Args:
            allow_std_error (bool): Whether to allow stderr in the output. Defaults to False.

        Returns:
            Callable[[Launcher], Self]: A callable that takes a launcher instance and returns the application instance.
        """

        def _run(launcher: Launcher):
            """Internal wrapper function"""
            try:
                self.add_app_settings(launcher=launcher)
                self.run()
                result = self.output_from_result(allow_stderr=allow_std_error)
            except subprocess.CalledProcessError as e:
                logger.error(f"App {self.__class__.__name__} failed with error: {e}")
                raise
            return result

        return _run


class AindBehaviorServicesBonsaiApp(BonsaiApp):
    """
    Specialized Bonsai application for AIND behavior services integration.

    This class extends the base BonsaiApp to provide specific functionality for
    AIND behavior experiments, including automatic configuration of task logic,
    session, and rig paths for the Bonsai workflow.

    Example:
        ```python
        # Create an AIND behavior services Bonsai app
        app = AindBehaviorServicesBonsaiApp(workflow="behavior_workflow.bonsai")
        app.add_app_settings(launcher=my_launcher)
        app.run()
        ```
    """

    def add_app_settings(self, *, launcher: Optional[Launcher] = None, **kwargs) -> Self:
        """
        Adds AIND behavior-specific application settings to the Bonsai workflow.

        This method automatically configures the TaskLogicPath, SessionPath, and RigPath
        properties for the Bonsai workflow based on the launcher's schema models.

        Args:
            launcher: The behavior launcher instance containing schema models
            **kwargs: Additional keyword arguments

        Returns:
            Self: The updated instance

        Raises:
            ValueError: If the required launcher argument is not provided

        Example:
            ```python
            # Add AIND behavior settings
            app.add_app_settings(launcher=my_launcher)
            ```
        """

        if launcher is None:
            raise ValueError("Missing required argument 'launcher'.")

        settings = {
            "TaskLogicPath": os.path.abspath(
                launcher.save_temp_model(model=launcher.get_task_logic(strict=True), directory=launcher.temp_dir)
            ),
            "SessionPath": os.path.abspath(
                launcher.save_temp_model(model=launcher.get_session(strict=True), directory=launcher.temp_dir)
            ),
            "RigPath": os.path.abspath(
                launcher.save_temp_model(model=launcher.get_rig(strict=True), directory=launcher.temp_dir)
            ),
        }
        return super().add_app_settings(**settings)



from unittest.mock import MagicMock, patch

import pytest

from clabe.ui import DefaultUIHelper


@pytest.fixture
def ui_helper():
    return DefaultUIHelper(print_func=MagicMock())


class TestDefaultUiHelper:
    @patch("builtins.input", side_effect=["Some notes"])
    def test_prompt_get_text(self, mock_input, ui_helper):
        result = ui_helper.prompt_text("")
        assert isinstance(result, str)

    @patch("builtins.input", side_effect=["Y"])
    def test_prompt_yes_no_question(self, mock_input, ui_helper):
        result = ui_helper.prompt_yes_no_question("Continue?")
        assert isinstance(result, bool)

    @patch("builtins.input", side_effect=["1"])
    def test_prompt_pick_from_list(self, mock_input, ui_helper):
        result = ui_helper.prompt_pick_from_list(["item1", "item2"], "Choose an item")
        assert isinstance(result, str)
        assert result == "item1"

    @patch("builtins.input", side_effect=["0"])
    def test_prompt_pick_from_list_none(self, mock_input, ui_helper):
        result = ui_helper.prompt_pick_from_list(["item1", "item2"], "Choose an item")
        assert result is None



import concurrent.futures
import getpass
import logging
import platform
from typing import Optional

logger = logging.getLogger(__name__)

if platform.system() == "Windows":
    import ldap3
    import ms_active_directory

    def validate_aind_username(
        username: str,
        domain: str = "corp.alleninstitute.org",
        domain_username: Optional[str] = None,
        timeout: Optional[float] = 2,
    ) -> bool:
        """
        Validates if the given username exists in the AIND Active Directory.

        This function authenticates with the corporate Active Directory and searches
        for the specified username to verify its existence within the organization.
        See https://github.com/AllenNeuralDynamics/aind-watchdog-service/issues/110#issuecomment-2828869619

        Args:
            username: The username to validate against Active Directory
            domain: The Active Directory domain to search. Defaults to Allen Institute domain
            domain_username: Username for domain authentication. Defaults to current user
            timeout: Timeout in seconds for the validation operation

        Returns:
            bool: True if the username exists in Active Directory, False otherwise

        Raises:
            concurrent.futures.TimeoutError: If the validation operation times out

        Example:
            ```python
            # Validate a username in Active Directory
            is_valid = validate_aind_username("j.doe")
            ```
        """

        def _helper(username: str, domain: str, domain_username: Optional[str]) -> bool:
            """A function submitted to a thread pool to validate the username."""
            if domain_username is None:
                domain_username = getpass.getuser()

            _domain = ms_active_directory.ADDomain(domain)
            session = _domain.create_session_as_user(
                domain_username,
                authentication_mechanism=ldap3.SASL,
                sasl_mechanism=ldap3.GSSAPI,
            )
            return session.find_user_by_name(username) is not None

        try:
            with concurrent.futures.ThreadPoolExecutor() as executor:
                future = executor.submit(_helper, username, domain, domain_username)
                result = future.result(timeout=timeout)
                return result
        except concurrent.futures.TimeoutError as e:
            logger.error("Timeout occurred while validating username: %s", e)
            e.add_note("Timeout occurred while validating username")
            raise e

else:

    def validate_aind_username(
        username: str,
        domain: str = "corp.alleninstitute.org",
        domain_username: Optional[str] = None,
        timeout: Optional[float] = 2,
    ) -> bool:
        """
        Validates if the given username is in the AIND Active Directory.

        This function is a no-op on non-Windows platforms since Active Directory
        authentication is not available.

        This function always returns True on non-Windows platforms.

        Args:
            username: The username to validate (ignored on non-Windows platforms)
            domain: The Active Directory domain (ignored on non-Windows platforms)
            domain_username: Username for domain authentication (ignored on non-Windows platforms)
            timeout: Timeout for the validation operation (ignored on non-Windows platforms)

        Returns:
            bool: Always returns True on non-Windows platforms
        """
        logger.warning("Active Directory validation is not implemented for non-Windows platforms")
        return True



import logging
import os
import subprocess
import typing as t
from pathlib import Path

import aind_behavior_curriculum.trainer
import pydantic

from ..launcher import Launcher
from ..launcher._callable_manager import Promise
from ..services import ServiceSettings
from ._base import App
from ._python_script import PythonScriptApp

if t.TYPE_CHECKING:
    from ..launcher import Launcher
else:
    Launcher = t.Any

P = t.ParamSpec("P")


logger = logging.getLogger(__name__)


class CurriculumSuggestion(pydantic.BaseModel):
    """
    Model representing a curriculum suggestion with trainer state and metrics.

    This model encapsulates the output from a curriculum run, including the updated
    trainer state, performance metrics, and version information.

    Attributes:
        trainer_state: The updated trainer state after curriculum processing
        metrics: Performance metrics from the curriculum run
        version: Version of the curriculum
        dsl_version: Version of the domain-specific language package used (aind-behavior-curriculum)
    """

    trainer_state: pydantic.SerializeAsAny[aind_behavior_curriculum.trainer.TrainerState]
    metrics: pydantic.SerializeAsAny[aind_behavior_curriculum.Metrics]
    version: str
    dsl_version: str


class CurriculumSettings(ServiceSettings):
    """
    Settings for the CurriculumApp.

    Attributes:
        entry_point: Path to the curriculum module (directory or file)
        input_trainer_state: Optional path to input TrainerState serialized file
        data_directory: Optional data directory for metrics calculation
    """

    __yml_section__: t.ClassVar[t.Literal["curriculum"]] = "curriculum"

    script: str = "curriculum run"
    project_directory: os.PathLike = Path(".")
    input_trainer_state: t.Optional[os.PathLike] = None
    data_directory: t.Optional[os.PathLike] = None
    curriculum: t.Optional[str] = None


class CurriculumApp(App):
    """
    A curriculum application that manages the execution of behavior curriculum scripts.

    This class facilitates running curriculum modules within a managed Python environment,
    handling trainer state input/output and data directory management for curriculum processing.

    Attributes:
        _settings: The curriculum settings configuration
        _python_script_app: Internal Python script application for execution

    Example:
        ```python
        # Create and run a curriculum app
        settings = CurriculumSettings(entry="/path/to/curriculum")
        app = CurriculumApp(settings)
        app.run()

        # Use with launcher for automated curriculum processing
        launcher.register_callable(app.build_runner(input_state_promise))
        ```
    """

    def __init__(self, settings: CurriculumSettings):
        """
        Initializes the CurriculumApp with the specified settings.

        Args:
            settings: Configuration settings for the curriculum application

        Raises:
            FileNotFoundError: If pyproject.toml cannot be found in parent directories

        Example:
            ```python
            settings = CurriculumSettings(
                entry_point="/path/to/curriculum/module",
                data_directory="/data/session"
            )
            app = CurriculumApp(settings)
            ```
        """
        self._settings = settings

        self._python_script_app = PythonScriptApp(
            script=settings.script, project_directory=settings.project_directory, extra_uv_arguments="-q"
        )

    def run(self) -> subprocess.CompletedProcess:
        """
        Executes the curriculum module with the configured settings.

        Returns:
            subprocess.CompletedProcess: The result of the curriculum execution

        Raises:
            ValueError: If input_trainer_state or data_directory is not set
            subprocess.CalledProcessError: If the curriculum script execution fails

        Example:
            ```python
            # Set required parameters and run
            app._settings.input_trainer_state = "/path/to/trainer_state.json"
            app._settings.data_directory = "/path/to/data"
            result = app.run()
            print(f"Exit code: {result.returncode}")
            ```
        """
        if self._settings.input_trainer_state is None:
            raise ValueError("Input trainer state is not set.")
        if self._settings.data_directory is None:
            raise ValueError("Data directory is not set.")

        kwargs: dict[str, t.Any] = {  # Must use kebab casing
            "data-directory": f'"{self._settings.data_directory}"',
            "input-trainer-state": f'"{self._settings.input_trainer_state}"',
        }
        if self._settings.curriculum is not None:
            kwargs["curriculum"] = f'"{self._settings.curriculum}"'

        self._python_script_app.add_app_settings(**kwargs)
        return self._python_script_app.run()

    def output_from_result(self, *, allow_stderr: bool | None = None) -> t.Self:
        """
        Processes the output from the curriculum execution result.

        Args:
            allow_stderr: Whether to allow stderr in the output. If None, uses default behavior

        Returns:
            Self: The current CurriculumApp instance

        Raises:
            subprocess.CalledProcessError: If the process failed or stderr is present when not allowed

        Example:
            ```python
            # Process output and handle errors
            try:
                app.output_from_result(allow_stderr=True)
                print("Curriculum completed successfully")
            except subprocess.CalledProcessError as e:
                print(f"Curriculum failed: {e}")
            ```
        """
        self._python_script_app.output_from_result(allow_stderr=allow_stderr)
        return self

    def add_app_settings(self, **kwargs) -> t.Self:
        """
        Adds application-specific settings to the curriculum execution.

        Args:
            **kwargs: Additional keyword arguments to pass to the curriculum script

        Returns:
            Self: The current CurriculumApp instance

        Example:
            ```python
            # Add custom settings
            app.add_app_settings(
                debug_mode=True,
                log_level="DEBUG",
                custom_param="value"
            )
            ```
        """
        self._python_script_app.add_app_settings(**kwargs)
        return self

    @property
    def result(self) -> subprocess.CompletedProcess:
        """
        Retrieves the result of the curriculum execution.

        Returns:
            subprocess.CompletedProcess: The result of the curriculum script execution

        Raises:
            RuntimeError: If the curriculum has not been run yet

        Example:
            ```python
            # Get execution result after running
            app.run()
            result = app.result
            print(f"Return code: {result.returncode}")
            print(f"Output: {result.stdout}")
            ```
        """
        return self._python_script_app.result

    def build_runner(
        self,
        input_trainer_state: Promise[P, aind_behavior_curriculum.trainer.TrainerState],
        *,
        session_directory: t.Optional[Promise[P, os.PathLike]] = None,
        allow_std_error: bool = False,
    ) -> t.Callable[[Launcher], CurriculumSuggestion]:
        """
        Builds a runner function for curriculum execution within a launcher context.

        This method creates a callable that can be registered with a launcher to run
        the curriculum with proper data directory and trainer state management.

        Args:
            input_trainer_state: Promise containing the trainer state to process
            allow_std_error: Whether to allow stderr output without raising an error

        Returns:
            Callable that takes a Launcher and returns a CurriculumSuggestion

        Raises:
            subprocess.CalledProcessError: If curriculum execution fails

        Note:
            This method overrides the base App.build_runner to provide curriculum-specific
            functionality and return type compatibility.

        Example:
            ```python
            # Register curriculum with launcher
            trainer_state_promise = launcher.register_callable(get_trainer_state)
            curriculum_runner = CurriculumApp.build_runner(trainer_state_promise, settings)
            launcher.register_callable(curriculum_runner)
            ```
        """

        def _run(launcher: Launcher) -> CurriculumSuggestion:
            if self._settings.data_directory is None:
                self._settings.data_directory = (
                    session_directory.result if session_directory else launcher.session_directory
                )
            if self._settings.input_trainer_state is None:
                self._settings.input_trainer_state = Path(launcher.save_temp_model(input_trainer_state.result))
            try:
                self.run()
                self.output_from_result(allow_stderr=allow_std_error)
            except subprocess.CalledProcessError as e:
                logger.error("App %s failed with error: %s", self.__class__.__name__, e)
                raise
            return self.get_suggestion()

        return _run

    def get_suggestion(self) -> CurriculumSuggestion:
        """
        Parses and returns the curriculum suggestion from the execution result.

        Returns:
            CurriculumSuggestion: Parsed curriculum output containing trainer state and metrics

        Raises:
            pydantic.ValidationError: If the result stdout cannot be parsed as valid JSON
            RuntimeError: If the curriculum has not been run yet

        Example:
            ```python
            # Get suggestion after running curriculum
            app.run()
            app.output_from_result()
            suggestion = app.get_suggestion()
            print(f"New trainer state: {suggestion.trainer_state}")
            print(f"Metrics: {suggestion.metrics}")
            ```
        """
        return CurriculumSuggestion.model_validate_json(self.result.stdout)



__version__ = "0.6.10"

import logging

from . import logging_helper

logger = logging.getLogger(__name__)

logging.basicConfig(
    level=logging.INFO,
    format=logging_helper.log_fmt,
    datefmt=logging_helper.datetime_fmt,
    handlers=[logging_helper.rich_handler],
)



import logging
from pathlib import Path
from unittest.mock import MagicMock, patch

from clabe.launcher import Launcher
from clabe.launcher._cli import LauncherCliArgs


def test_base_launcher_init_basic(
    mock_base_launcher,
    mock_rig,
    mock_session,
    mock_task_logic,
):
    """Test basic initialization of Launcher."""
    assert mock_base_launcher.get_rig() == mock_rig
    assert mock_base_launcher.get_session() == mock_session
    assert mock_base_launcher.get_task_logic() == mock_task_logic
    assert mock_base_launcher.get_rig_model() is type(mock_rig)
    assert mock_base_launcher.get_session_model() is type(mock_session)
    assert mock_base_launcher.get_task_logic_model() is type(mock_task_logic)
    assert Path(mock_base_launcher.settings.temp_dir).exists()


def test_base_launcher_with_attached_logger(
    mock_base_launcher, mock_rig, mock_session, mock_task_logic, mock_ui_helper
):
    """Test launcher initialization with attached logger."""
    with patch("clabe.logging_helper.add_file_handler") as mock_add_file_handler:
        mock_attached_logger = MagicMock()
        launcher = Launcher(
            rig=mock_rig,
            session=mock_session,
            task_logic=mock_task_logic,
            ui_helper=mock_ui_helper,
            settings=mock_base_launcher.settings,
            attached_logger=mock_attached_logger,
        )
        assert launcher.logger == mock_add_file_handler.return_value
        mock_add_file_handler.assert_called()


def test_base_launcher_debug_mode(mock_rig, mock_session, mock_task_logic, mock_ui_helper, tmp_path: Path):
    """Test launcher initialization with debug mode enabled."""
    launcher_args_debug = LauncherCliArgs(
        data_dir=tmp_path / "data",
        temp_dir=tmp_path / "temp",
        debug_mode=True,
        create_directories=True,
    )
    with patch("clabe.launcher._base.GitRepository") as mock_git, patch("os.chdir"), patch("pathlib.Path.mkdir"):
        mock_git.return_value.working_dir = launcher_args_debug.data_dir
        with patch("clabe.logging_helper.add_file_handler") as mock_add_file_handler:
            mock_logger = MagicMock()
            mock_add_file_handler.return_value = mock_logger
            Launcher(
                rig=mock_rig,
                session=mock_session,
                task_logic=mock_task_logic,
                ui_helper=mock_ui_helper,
                settings=launcher_args_debug,
            )
            mock_logger.setLevel.assert_called_with(logging.DEBUG)


def test_base_launcher_create_directories(mock_rig, mock_session, mock_task_logic, mock_ui_helper, tmp_path: Path):
    """Test launcher initialization with create_directories option."""
    launcher_args_create_dirs = LauncherCliArgs(
        data_dir=tmp_path / "data",
        temp_dir=tmp_path / "temp",
        create_directories=True,
    )
    with (
        patch("clabe.launcher._base.GitRepository") as mock_git,
        patch("os.chdir"),
        patch("pathlib.Path.mkdir"),
        patch("clabe.logging_helper.add_file_handler") as log_mod,
    ):
        log_mod.return_value = MagicMock()
        mock_git.return_value.working_dir = launcher_args_create_dirs.data_dir
        with patch("clabe.launcher.Launcher._create_directory_structure") as mock_create_dirs:
            Launcher(
                rig=mock_rig,
                session=mock_session,
                task_logic=mock_task_logic,
                ui_helper=mock_ui_helper,
                settings=launcher_args_create_dirs,
                attached_logger=log_mod.return_value,
            )
            mock_create_dirs.assert_called_once()


def test_create_directory():
    with patch("os.makedirs") as mock_makedirs, patch("os.path.exists", return_value=False):
        directory = Path("/tmp/fake/directory")
        Launcher.create_directory(directory)
        mock_makedirs.assert_called_once_with(directory)


def test_create_directory_structure(mock_rig, mock_session, mock_task_logic, mock_ui_helper, tmp_path: Path):
    """Test that _create_directory_structure calls create_directory for data_dir and temp_dir."""
    launcher_args = LauncherCliArgs(
        data_dir=tmp_path / "data",
        temp_dir=tmp_path / "temp",
        create_directories=True,
    )
    with (
        patch("clabe.launcher._base.GitRepository") as mock_git,
        patch("os.chdir"),
        patch("pathlib.Path.mkdir"),
        patch("clabe.logging_helper.add_file_handler") as log_mod,
    ):
        mock_git.return_value.working_dir = launcher_args.data_dir
        log_mod.return_value = MagicMock()
        with patch("clabe.launcher.Launcher.create_directory") as mock_create_directory:
            launcher_args.create_directories = True
            launcher = Launcher(
                rig=mock_rig,
                session=mock_session,
                task_logic=mock_task_logic,
                ui_helper=mock_ui_helper,
                settings=launcher_args,
                attached_logger=log_mod.return_value,
            )
            mock_create_directory.assert_any_call(launcher.settings.data_dir)
            mock_create_directory.assert_any_call(launcher.temp_dir)
            assert mock_create_directory.call_count == 2



import subprocess
from pathlib import Path
from unittest.mock import MagicMock

import pytest

from clabe.git_manager._git import GitRepository


@pytest.fixture
def temp_git_repo(tmp_path: Path) -> Path:
    """Create a temporary git repository."""
    git_dir = tmp_path / "test_repo"
    git_dir.mkdir()
    subprocess.run(["git", "init"], cwd=git_dir, check=True)
    (git_dir / "test_file.txt").write_text("hello")
    subprocess.run(["git", "add", "test_file.txt"], cwd=git_dir, check=True)
    try:
        subprocess.run(
            ["git", "commit", "-m", "initial commit"],
            cwd=git_dir,
            check=True,
            capture_output=True,
        )
    except subprocess.CalledProcessError as e:
        # It's possible git is not configured with user.name and user.email
        if b"Please tell me who you are" in e.stderr:
            subprocess.run(
                ["git", "config", "user.email", "test@example.com"],
                cwd=git_dir,
                check=True,
            )
            subprocess.run(
                ["git", "config", "user.name", "Test User"],
                cwd=git_dir,
                check=True,
            )
            subprocess.run(
                ["git", "commit", "-m", "initial commit"],
                cwd=git_dir,
                check=True,
            )
        else:
            raise e
    return git_dir


def test_git_repository_init(temp_git_repo: Path):
    """Test GitRepository initialization."""
    repo = GitRepository(path=temp_git_repo)
    assert repo.working_dir == str(temp_git_repo)


def test_reset_repo(temp_git_repo: Path):
    """Test reset_repo."""
    (temp_git_repo / "test_file.txt").write_text("world")
    repo = GitRepository(path=temp_git_repo)
    assert repo.is_dirty()
    repo.reset_repo()
    assert not repo.is_dirty()


def test_clean_repo(temp_git_repo: Path):
    """Test clean_repo."""
    (temp_git_repo / "untracked_file.txt").write_text("untracked")
    repo = GitRepository(path=temp_git_repo)
    assert "untracked_file.txt" in repo.untracked_files
    repo.clean_repo()
    assert "untracked_file.txt" not in repo.untracked_files


def test_is_dirty_with_submodules_clean(temp_git_repo: Path):
    """Test is_dirty_with_submodules on a clean repo."""
    repo = GitRepository(path=temp_git_repo)
    assert not repo.is_dirty_with_submodules()


def test_is_dirty_with_submodules_dirty_main(temp_git_repo: Path):
    """Test is_dirty_with_submodules on a dirty main repo."""
    (temp_git_repo / "test_file.txt").write_text("world")
    repo = GitRepository(path=temp_git_repo)
    assert repo.is_dirty_with_submodules()


def test_uncommitted_changes_clean(temp_git_repo: Path):
    """Test uncommitted_changes on a clean repo."""
    repo = GitRepository(path=temp_git_repo)
    assert repo.uncommitted_changes() == []


def test_uncommitted_changes_dirty(temp_git_repo: Path):
    """Test uncommitted_changes on a dirty repo."""
    (temp_git_repo / "test_file.txt").write_text("world")
    (temp_git_repo / "untracked_file.txt").write_text("untracked")
    repo = GitRepository(path=temp_git_repo)
    changes = repo.uncommitted_changes()
    assert "test_file.txt" in changes
    assert "untracked_file.txt" in changes


def test_full_reset(temp_git_repo: Path):
    """Test full_reset."""
    (temp_git_repo / "test_file.txt").write_text("world")
    (temp_git_repo / "untracked_file.txt").write_text("untracked")
    repo = GitRepository(path=temp_git_repo)
    repo.full_reset()
    assert not repo.is_dirty()
    assert not repo.untracked_files


def test_try_prompt_full_reset_no_reset(temp_git_repo: Path):
    """Test try_prompt_full_reset when user says no."""
    (temp_git_repo / "test_file.txt").write_text("world")
    repo = GitRepository(path=temp_git_repo)
    mock_ui = MagicMock()
    mock_ui.prompt_yes_no_question.return_value = False
    repo.try_prompt_full_reset(mock_ui)
    assert repo.is_dirty()


def test_try_prompt_full_reset_yes_reset(temp_git_repo: Path):
    """Test try_prompt_full_reset when user says yes."""
    (temp_git_repo / "test_file.txt").write_text("world")
    repo = GitRepository(path=temp_git_repo)
    mock_ui = MagicMock()
    mock_ui.prompt_yes_no_question.return_value = True
    repo.try_prompt_full_reset(mock_ui)
    assert not repo.is_dirty()


def test_try_prompt_full_reset_force_reset(temp_git_repo: Path):
    """Test try_prompt_full_reset with force_reset=True."""
    (temp_git_repo / "test_file.txt").write_text("world")
    repo = GitRepository(path=temp_git_repo)
    mock_ui = MagicMock()
    repo.try_prompt_full_reset(mock_ui, force_reset=True)
    assert not repo.is_dirty()
    mock_ui.prompt_yes_no_question.assert_not_called()



import logging
import warnings
from pathlib import Path
from shutil import _ntuple_diskusage
from unittest.mock import MagicMock, patch

import pytest

from clabe.resource_monitor import (
    Constraint,
    ResourceMonitor,
    available_storage_constraint_factory,
    remote_dir_exists_constraint_factory,
)


@pytest.fixture
def monitor():
    warnings.simplefilter("ignore")
    logging.disable(logging.CRITICAL)
    return ResourceMonitor()


class TestResourceMonitor:
    def test_add_constraint(self, monitor):
        constraint = MagicMock(spec=Constraint)
        monitor.add_constraint(constraint)
        assert constraint in monitor.constraints

    def test_remove_constraint(self, monitor):
        constraint = MagicMock(spec=Constraint)
        monitor.add_constraint(constraint)
        monitor.remove_constraint(constraint)
        assert constraint not in monitor.constraints

    def test_evaluate_constraints_all_pass(self, monitor):
        constraint = MagicMock(spec=Constraint)
        constraint.return_value = True
        monitor.add_constraint(constraint)
        assert monitor.evaluate_constraints()

    def test_evaluate_constraints_one_fails(self, monitor):
        constraint1 = MagicMock(spec=Constraint)
        constraint1.return_value = True
        constraint2 = MagicMock(spec=Constraint)
        constraint2.return_value = False
        constraint2.on_fail.return_value = "Constraint failed"
        monitor.add_constraint(constraint1)
        monitor.add_constraint(constraint2)
        assert not monitor.evaluate_constraints()

    @patch("shutil.disk_usage")
    def test_available_storage_constraint_factory(self, mock_disk_usage):
        mock_disk_usage.return_value = _ntuple_diskusage(total=int(500e9), used=int(100e9), free=int(400e9))
        constraint = available_storage_constraint_factory(drive=Path("C:\\"), min_bytes=2e11)
        assert constraint()
        constraint = available_storage_constraint_factory(drive=Path("C:\\"), min_bytes=2e13)
        assert not constraint()

    @patch("os.path.exists")
    def test_remote_dir_exists_constraint_factory(self, mock_exists):
        mock_exists.return_value = True
        constraint = remote_dir_exists_constraint_factory(dir_path=Path("/some/remote/dir"))
        assert constraint()

    def test_resource_monitor_service(self):
        resource_monitor = ResourceMonitor()

        resource_monitor.add_constraint(
            Constraint(name="test_constraint", constraint=lambda: True, fail_msg_handler=lambda: "Constraint failed.")
        )

        assert resource_monitor.evaluate_constraints()

        resource_monitor.add_constraint(
            Constraint(name="test_constraint", constraint=lambda: False, fail_msg_handler=lambda: "Constraint failed.")
        )

        assert not resource_monitor.evaluate_constraints()

    def test_resource_monitor_service_constraint(self):
        constraint = Constraint(
            name="test_constraint", constraint=lambda x: x, fail_msg_handler=lambda: "Constraint failed.", args=[True]
        )

        assert constraint()

        constraint = Constraint(
            name="test_constraint", constraint=lambda x: x, fail_msg_handler=lambda: "Constraint failed.", args=[False]
        )
        assert not constraint()



"""Constants used throughout the CLABE package."""

import enum
import os
import typing as t
from pathlib import Path

from aind_behavior_services.utils import format_datetime, model_from_json_file, utcnow

__all__ = [
    "abspath",
    "format_datetime",
    "model_from_json_file",
    "utcnow",
    "ByAnimalFiles",
]

KNOWN_CONFIG_FILES: t.List[str] = [
    "./clabe.yml",
    "./local/clabe.yml",
    "./.local/clabe.yml",
]


def abspath(path: os.PathLike) -> Path:
    """
    Helper method that converts a path to an absolute path.

    Args:
        path: The path to convert

    Returns:
        Path: The absolute path
    """
    return Path(path).resolve()


class ByAnimalFiles(enum.StrEnum):
    """
    Enum for file types associated with animals in the experiment.

    Defines the standard file types that can be associated with individual
    animals/subjects in behavior experiments.

    Example:
        ```python
        # Use the task logic file type
        filename = f"{ByAnimalFiles.TASK_LOGIC}.json"
        ```
    """

    TASK_LOGIC = "task_logic"
    TRAINER_STATE = "trainer_state"



from pathlib import Path

import pytest

from clabe.apps import CurriculumApp, CurriculumSettings, CurriculumSuggestion, PythonScriptApp

from .. import TESTS_ASSETS, SubmoduleManager

SubmoduleManager.initialize_submodules()


@pytest.fixture
def curriculum_app() -> CurriculumApp:
    """Fixture to create a CurriculumApp for the curriculum tests."""

    return CurriculumApp(
        settings=CurriculumSettings(
            script="curriculum run",
            input_trainer_state=Path("MockPath"),
            data_directory="Demo",
            project_directory=TESTS_ASSETS / "Aind.Behavior.VrForaging.Curricula",
            curriculum="template",
        )
    )


@pytest.fixture
def inner_python_app(curriculum_app: CurriculumApp):
    return curriculum_app._python_script_app


class TestCurriculumIntegration:
    """Tests the integration with the aind-behavior-curriculum submodule."""

    def test_can_create_venv(self, inner_python_app: PythonScriptApp) -> None:
        """Tests that the virtual environment can be created."""
        proc = inner_python_app.create_environment()
        proc.check_returncode()

    def test_curriculum_run(self, curriculum_app: CurriculumApp) -> None:
        """Tests that the curriculum can be run."""

        curriculum_app.run()
        curriculum_app.result.check_returncode()
        parsed_output = CurriculumSuggestion.model_validate_json(curriculum_app.output_from_result().result.stdout)
        with open(TESTS_ASSETS / "expected_curriculum_suggestion.json", "r", encoding="utf-8") as f:
            expected = f.read()
            assert parsed_output == CurriculumSuggestion.model_validate_json(expected)



import importlib.util

if importlib.util.find_spec("aind_data_transfer_service") is None:
    raise ImportError(
        "The 'aind_data_transfer_service' package is required to use this module. \
            Install the optional dependencies defined in `project.toml' \
                by running `pip install .[aind-services]`"
    )

import datetime
import json
import logging
import os
import subprocess
from os import PathLike
from pathlib import Path, PurePosixPath
from typing import TYPE_CHECKING, Any, Callable, ClassVar, Dict, Generic, List, Optional, TypeVar, Union

import aind_data_transfer_service.models.core
import pydantic
import requests
import yaml
from aind_data_schema.core.metadata import CORE_FILES
from pydantic import BaseModel, SerializeAsAny, TypeAdapter
from requests.exceptions import HTTPError

from .. import ui
from ..launcher._callable_manager import Promise
from ..services import ServiceSettings
from ._aind_watchdog_models import (
    DEFAULT_TRANSFER_ENDPOINT,
    BucketType,
    ManifestConfig,
    Platform,
    WatchConfig,
)
from ._base import DataTransfer

if TYPE_CHECKING:
    from aind_data_schema.core.session import Session as AdsSession

    from ..data_mapper.aind_data_schema import AindDataSchemaSessionDataMapper
    from ..launcher import Launcher
else:
    Launcher = Any
    AdsSession = Any
    AindDataSchemaSessionDataMapper = Any

logger = logging.getLogger(__name__)

TransferServiceTask = Dict[
    str, Union[aind_data_transfer_service.models.core.Task, Dict[str, aind_data_transfer_service.models.core.Task]]
]
TSessionMapper = TypeVar("TSessionMapper", bound=AindDataSchemaSessionDataMapper)


class WatchdogSettings(ServiceSettings):
    """
    Settings for the WatchdogDataTransferService.

    Attributes:
        destination (Path): The destination path for the data transfer.
        schedule_time (Optional[datetime.time]): The time to schedule the data transfer.
        project_name (str): The name of the project.
        platform (Platform): The platform of the project.
        capsule_id (Optional[str]): The capsule ID for the data transfer.
        script (Optional[Dict[str, List[str]]]): A dictionary of scripts to run.
        s3_bucket (BucketType): The S3 bucket to transfer the data to.
        mount (Optional[str]): The mount point for the data transfer.
        force_cloud_sync (bool): Whether to force a cloud sync.
        transfer_endpoint (str): The endpoint for the data transfer service.
        delete_modalities_source_after_success (bool): Whether to delete the source data after a successful transfer.
        extra_identifying_info (Optional[dict]): Extra identifying information for the data transfer.
        upload_tasks (Optional[Any]): Upload job configurations. Use the placeholder "{{ destination }}" to later reference the destination path.
        job_config (str): Job configuration name.
    """

    __yml_section__: ClassVar[Optional[str]] = "watchdog"

    destination: Path
    schedule_time: Optional[datetime.time] = datetime.time(hour=20)
    project_name: str
    platform: Platform = "behavior"
    capsule_id: Optional[str] = None
    script: Optional[Dict[str, List[str]]] = None
    s3_bucket: BucketType = "private"
    mount: Optional[str] = None
    force_cloud_sync: bool = True
    transfer_endpoint: str = DEFAULT_TRANSFER_ENDPOINT
    delete_modalities_source_after_success: bool = False
    extra_identifying_info: Optional[dict] = None
    upload_tasks: Optional[SerializeAsAny[TransferServiceTask]] = None
    job_type: str = "default"


class WatchdogDataTransferService(DataTransfer[WatchdogSettings], Generic[TSessionMapper]):
    """
    A data transfer service that uses the aind-watchdog-service to monitor and transfer
    data based on manifest configurations.

    This service integrates with the AIND data transfer infrastructure to automatically
    monitor directories for new data and transfer it to specified destinations with
    proper metadata handling and validation.

    Attributes:
        _source (PathLike): Source directory to monitor
        _settings (WatchdogSettings): Service settings containing destination and configuration
        _aind_session_data_mapper (Optional[_TSessionMapper]): Mapper for session data
        _ui_helper (ui.UiHelper): UI helper for user prompts
        Various configuration attributes accessible via settings

    Example:
        ```python
        # Basic watchdog service setup:
        settings = WatchdogSettings(
            destination="//server/data/session_001",
            project_name="my_project"
        )
        service = WatchdogDataTransferService(
            source="C:/data/session_001",
            settings=settings
        )

        # Full configuration with session mapper:
        settings = WatchdogSettings(
            destination="//server/data/session_001",
            project_name="behavior_study",
            schedule_time=datetime.time(hour=22, minute=30),
            platform=Platform.BEHAVIOR,
            force_cloud_sync=True
        )
        session_mapper = MySessionMapper(session_data)
        service = WatchdogDataTransferService(
            source="C:/data/session_001",
            settings=settings
        )
        service = service.with_aind_session_data_mapper(session_mapper)
        if service.validate():
            service.transfer()
        ```
    """

    def __init__(
        self,
        source: PathLike,
        settings: WatchdogSettings,
        *,
        validate: bool = True,
        session_name: Optional[str] = None,
        ui_helper: Optional[ui.UiHelper] = None,
    ) -> None:
        """
        Initializes the WatchdogDataTransferService.

        Args:
            source: The source directory or file to monitor
            settings: WatchdogSettings containing destination and configuration options
            validate: Whether to validate the project name
            session_name: Name of the session
            ui_helper: UI helper for user prompts

        Example:
            ```python
            # Basic initialization:
            settings = WatchdogSettings(
                destination="//server/archive/session_001",
                project_name="behavior_project"
            )
            service = WatchdogDataTransferService(
                source="C:/data/session_001",
                settings=settings
            )

            # Advanced configuration:
            settings = WatchdogSettings(
                destination="//server/archive/session_001",
                project_name="behavior_project",
                schedule_time=datetime.time(hour=23),
                platform=Platform.BEHAVIOR,
                force_cloud_sync=True,
                delete_modalities_source_after_success=True,
                extra_identifying_info={"experiment_type": "foraging"}
            )
            service = WatchdogDataTransferService(
                source="C:/data/session_001",
                settings=settings
            )
            ```
        """
        self._settings = settings
        self._source = source

        self._aind_session_data_mapper: Optional[TSessionMapper] = None

        _default_exe = os.environ.get("WATCHDOG_EXE", None)
        _default_config = os.environ.get("WATCHDOG_CONFIG", None)

        if _default_exe is None or _default_config is None:
            raise ValueError("WATCHDOG_EXE and WATCHDOG_CONFIG environment variables must be defined.")

        self.executable_path = Path(_default_exe)
        self.config_path = Path(_default_config)

        self._watch_config: Optional[WatchConfig] = None
        self._manifest_config: Optional[ManifestConfig] = None

        self._validate_project_name = validate
        self._ui_helper = ui_helper or ui.DefaultUIHelper()
        self._session_name = session_name

    @property
    def aind_session_data_mapper(self) -> TSessionMapper:
        """
        Gets the aind-data-schema session data mapper.

        Returns:
            The session data mapper

        Raises:
            ValueError: If the data mapper is not set
        """
        if self._aind_session_data_mapper is None:
            raise ValueError("Data mapper is not set.")
        return self._aind_session_data_mapper

    def with_aind_session_data_mapper(self, value: TSessionMapper) -> "WatchdogDataTransferService[TSessionMapper]":
        """
        Sets the aind-data-schema session data mapper.

        Args:
            value: The data mapper to set

        Raises:
            ValueError: If the provided value is not a valid data mapper
        """
        self._aind_session_data_mapper = value
        return self

    def transfer(self) -> None:
        """
        Executes the data transfer by generating a Watchdog manifest configuration.

        Creates and deploys a manifest configuration file that the watchdog service
        will use to monitor and transfer data according to the specified parameters.
        """
        try:
            if not self.is_running():
                logger.warning("Watchdog service is not running. Attempting to start it.")
                try:
                    self.force_restart(kill_if_running=False)
                except subprocess.CalledProcessError as e:
                    logger.error("Failed to start watchdog service. %s", e)
                    raise RuntimeError("Failed to start watchdog service.") from e
                else:
                    if not self.is_running():
                        logger.error("Failed to start watchdog service.")
                        raise RuntimeError("Failed to start watchdog service.")
                    else:
                        logger.info("Watchdog service restarted successfully.")

            logger.info("Creating watchdog manifest config.")

            if not self.aind_session_data_mapper.is_mapped():
                raise ValueError("Data mapper has not been mapped yet.")

            self._manifest_config = self._create_manifest_config_from_ads_session(
                ads_session=self.aind_session_data_mapper.mapped,
                session_name=self._session_name,
            )

            if self._watch_config is None:
                raise ValueError("Watchdog config is not set.")

            assert self._manifest_config.name is not None, "Manifest config name must be set."
            _manifest_path = self.dump_manifest_config(
                path=Path(self._watch_config.flag_dir) / self._manifest_config.name
            )
            logger.info("Watchdog manifest config created successfully at %s.", _manifest_path)

        except (pydantic.ValidationError, ValueError, IOError) as e:
            logger.error("Failed to create watchdog manifest config. %s", e)
            raise e

    def validate(self) -> bool:
        """
        Validates the Watchdog service and its configuration.

        Checks for required executables, configuration files, service status,
        and project name validity.

        Returns:
            True if the service is valid, False otherwise

        Raises:
            FileNotFoundError: If required files are missing
            HTTPError: If the project name validation fails
        """
        logger.info("Attempting to validate Watchdog service.")
        if not self.executable_path.exists():
            raise FileNotFoundError(f"Executable not found at {self.executable_path}")
        if not self.config_path.exists():
            raise FileNotFoundError(f"Config file not found at {self.config_path}")
        self._watch_config = WatchConfig.model_validate(self._read_yaml(self.config_path))

        if not self.is_running():
            logger.warning(
                "Watchdog service is not running. \
                                After the session is over, \
                                the launcher will attempt to forcefully restart it"
            )
            return False

        try:
            _valid_proj = self.is_valid_project_name()
            if not _valid_proj:
                logger.warning("Watchdog project name is not valid.")
        except HTTPError as e:
            logger.error("Failed to fetch project names from endpoint. %s", e)
            raise e
        return _valid_proj

    def is_valid_project_name(self) -> bool:
        """
        Checks if the project name is valid by querying the metadata service.

        Validates the project name against the list of known projects from
        the AIND metadata service.

        Returns:
            True if the project name is valid, False otherwise
        """
        project_names = self._get_project_names()
        return self._settings.project_name in project_names

    def _create_manifest_config_from_ads_session(
        self,
        ads_session: AdsSession,
        ads_schemas: Optional[List[os.PathLike]] = None,
        session_name: Optional[str] = None,
    ) -> ManifestConfig:
        """
        Creates a ManifestConfig object from an aind-data-schema session.

        Converts session metadata into a manifest configuration that can be
        used by the watchdog service for data transfer operations.

        Args:
            ads_session: The aind-data-schema session data
            ads_schemas: Optional list of schema files
            session_name: Name of the session

        Returns:
            A ManifestConfig object

        Raises:
            ValueError: If the project name is invalid

        Example:
            ```python
            # Create manifest from session data:
            session = Session(...)
            manifest = service.create_manifest_config_from_ads_session(
                ads_session=session,
            )

            # Create with custom schemas:
            schemas = ["C:/data/rig.json", "C:/data/processing.json"]
            manifest = service.create_manifest_config_from_ads_session(
                ads_session=session,
                ads_schemas=schemas,
            )
            ```
        """
        processor_full_name = ",".join(ads_session.experimenter_full_name) or os.environ.get("USERNAME", "unknown")

        destination = Path(self._settings.destination).resolve()
        source = Path(self._source).resolve()

        if self._validate_project_name:
            project_names = self._get_project_names()
            if self._settings.project_name not in project_names:
                raise ValueError(f"Project name {self._settings.project_name} not found in {project_names}")

        ads_schemas = self._find_ads_schemas(source) if ads_schemas is None else ads_schemas

        _manifest_config = ManifestConfig(
            name=session_name,
            modalities={
                str(modality.abbreviation): [Path(path.resolve()) for path in [source / str(modality.abbreviation)]]
                for modality in ads_session.data_streams[0].stream_modalities
            },
            subject_id=int(ads_session.subject_id),
            acquisition_datetime=ads_session.session_start_time,
            schemas=[Path(value) for value in ads_schemas],
            destination=Path(destination),
            mount=self._settings.mount,
            processor_full_name=processor_full_name,
            project_name=self._settings.project_name,
            schedule_time=self._settings.schedule_time,
            platform=self._settings.platform,
            capsule_id=self._settings.capsule_id,
            s3_bucket=self._settings.s3_bucket,
            script=self._settings.script if self._settings.script else {},
            force_cloud_sync=self._settings.force_cloud_sync,
            transfer_endpoint=self._settings.transfer_endpoint,
            delete_modalities_source_after_success=self._settings.delete_modalities_source_after_success,
            extra_identifying_info=self._settings.extra_identifying_info,
        )

        _manifest_config = self._make_transfer_args(
            _manifest_config,
            add_default_tasks=True,
            extra_tasks=self._settings.upload_tasks or {},
            job_type=self._settings.job_type,
        )
        return _manifest_config

    @staticmethod
    def _remote_destination_root(manifest: ManifestConfig) -> Path:
        """Determines the remote destination root path for the manifest."""
        assert manifest.destination is not None, "Manifest destination must be set."
        assert manifest.name is not None, "Manifest name must be set."
        return Path(manifest.destination) / manifest.name

    @classmethod
    def _make_transfer_args(
        cls,
        manifest: ManifestConfig,
        *,
        job_type: str = "default",
        add_default_tasks: bool = True,
        extra_tasks: TransferServiceTask,
    ) -> ManifestConfig:
        """Appends extra tasks to a manifest configuration. Additionally appends default
        metadata and modality transformation tasks if requested."""
        tasks = {}

        if add_default_tasks:
            tasks["modality_transformation_settings"] = {
                modality: aind_data_transfer_service.models.core.Task(
                    job_settings={"input_source": str(PurePosixPath(cls._remote_destination_root(manifest) / modality))}
                )
                for modality in manifest.modalities.keys()
            }

            tasks["gather_preliminary_metadata"] = aind_data_transfer_service.models.core.Task(
                job_settings={"metadata_dir": str(PurePosixPath(cls._remote_destination_root(manifest) / "metadata"))}
            )

        extra_tasks = cls._interpolate_from_manifest(
            extra_tasks,
            str(PurePosixPath(cls._remote_destination_root(manifest))),
            "{{ destination }}",
        )

        tasks.update(extra_tasks)

        upload_job_configs_v2 = aind_data_transfer_service.models.core.UploadJobConfigsV2(
            job_type=job_type,
            project_name=manifest.project_name,
            platform=aind_data_transfer_service.models.core.Platform.from_abbreviation(manifest.platform),
            modalities=[
                aind_data_transfer_service.models.core.Modality.from_abbreviation(m) for m in manifest.modalities.keys()
            ],
            subject_id=str(manifest.subject_id),
            acq_datetime=manifest.acquisition_datetime.strftime("%Y-%m-%d %H:%M:%S"),
            tasks=tasks,
            s3_bucket=manifest.s3_bucket,
            acquisition_datetime=manifest.acquisition_datetime,
        )

        submit_request_v2 = aind_data_transfer_service.models.core.SubmitJobRequestV2(
            upload_jobs=[upload_job_configs_v2],
        )
        manifest.transfer_service_args = submit_request_v2
        return manifest

    @staticmethod
    def _interpolate_from_manifest(
        tasks: TransferServiceTask | dict, value: str, placeholder: str
    ) -> TransferServiceTask:
        """
        Interpolates values from the manifest into the upload job configuration.

        Args:
            upload_job_configs: The upload job configuration to update
            destination: The destination path to use for interpolation
            placeholder: The placeholder string to replace
        """
        _adapter = TypeAdapter(TransferServiceTask)
        literal = _adapter.dump_json(tasks, serialize_as_any=True)
        updated_literal = literal.decode("utf-8").replace(placeholder, value)
        return _adapter.validate_json(updated_literal)

    @staticmethod
    def _find_ads_schemas(source: PathLike) -> List[PathLike]:
        """
        Finds aind-data-schema schema files in the source directory.

        Searches for standard AIND data schema files in the specified directory.

        Args:
            source: The source directory to search

        Returns:
            A list of schema file paths
        """
        json_files = []
        for core_file in CORE_FILES:
            json_file = Path(source) / f"{core_file}.json"
            if json_file.exists():
                json_files.append(json_file)
        return [path for path in json_files]

    @staticmethod
    def _get_project_names(
        end_point: str = "http://aind-metadata-service/project_names", timeout: int = 5
    ) -> list[str]:
        """
        Fetches the list of valid project names from the metadata service.

        Queries the AIND metadata service to retrieve the current list of
        valid project names for validation purposes.

        Args:
            end_point: The endpoint URL for the metadata service
            timeout: Timeout for the request

        Returns:
            A list of valid project names

        Raises:
            HTTPError: If the request fails
        """
        response = requests.get(end_point, timeout=timeout)
        if response.ok:
            return json.loads(response.content)["data"]
        else:
            response.raise_for_status()
            raise HTTPError(f"Failed to fetch project names from endpoint. {response.content.decode('utf-8')}")

    def is_running(self) -> bool:
        """
        Checks if the Watchdog service is currently running.

        Uses system process monitoring to determine if the watchdog executable
        is currently active.

        Returns:
            True if the service is running, False otherwise

        Example:
            ```python
            # Check service status:
            settings = WatchdogSettings(
                destination="//server/data",
                project_name="my_project"
            )
            service = WatchdogDataTransferService(source="C:/data", settings=settings)
            if service.is_running():
                print("Watchdog service is active")
            else:
                print("Watchdog service is not running")
                service.force_restart()
            ```
        """
        output = subprocess.check_output(
            ["tasklist", "/FI", f"IMAGENAME eq {self.executable_path.name}"], shell=True, encoding="utf-8"
        )
        processes = [line.split()[0] for line in output.splitlines()[2:]]
        return len(processes) > 0

    def force_restart(self, kill_if_running: bool = True) -> subprocess.Popen[bytes]:
        """
        Attempts to restart the Watchdog application.

        Terminates the existing service if running and starts a new instance
        with the current configuration.

        Args:
            kill_if_running: Whether to terminate the service if it's already running

        Returns:
            A subprocess.Popen object representing the restarted service
        """
        if kill_if_running is True:
            while self.is_running():
                subprocess.run(["taskkill", "/IM", self.executable_path.name, "/F"], shell=True, check=True)

        cmd_factory = "{exe} -c {config}".format(exe=self.executable_path, config=self.config_path)

        return subprocess.Popen(cmd_factory, start_new_session=True, shell=True)

    def dump_manifest_config(self, path: Optional[os.PathLike] = None, make_dir: bool = True) -> Path:
        """
        Dumps the manifest configuration to a YAML file.

        Saves the current manifest configuration to a file that can be
        processed by the watchdog service.

        Args:
            path: The file path to save the manifest
            make_dir: Whether to create the directory if it doesn't exist

        Returns:
            The path to the saved manifest file

        Raises:
            ValueError: If the manifest or watch configuration is not set
        """
        manifest_config = self._manifest_config
        watch_config = self._watch_config

        if manifest_config is None or watch_config is None:
            raise ValueError("ManifestConfig or WatchConfig config is not set.")

        path = (Path(path) if path else Path(watch_config.flag_dir) / f"manifest_{manifest_config.name}.yaml").resolve()
        if not path.name.startswith("manifest_"):
            logger.info("Prefix manifest_ not found in file name. Appending it.")
            path = path.with_name(f"manifest_{path.stem}{path.suffix}")

        if make_dir and not path.parent.exists():
            path.parent.mkdir(parents=True, exist_ok=True)

        manifest_config.destination = Path(manifest_config.destination)
        manifest_config.schemas = [Path(schema) for schema in manifest_config.schemas]
        for modality in manifest_config.modalities:
            manifest_config.modalities[modality] = [_path for _path in manifest_config.modalities[modality]]

        self._write_yaml(manifest_config, path)
        return path

    @staticmethod
    def _yaml_dump(model: BaseModel) -> str:
        """
        Converts a Pydantic model to a YAML string.

        Serializes a Pydantic model to YAML format for file output.

        Args:
            model: The Pydantic model to convert

        Returns:
            A YAML string representation of the model
        """
        native_json = json.loads(model.model_dump_json())
        return yaml.dump(native_json, default_flow_style=False)

    @classmethod
    def _write_yaml(cls, model: BaseModel, path: PathLike) -> None:
        """
        Writes a Pydantic model to a YAML file.

        Saves a Pydantic model as a YAML file at the specified path.

        Args:
            model: The Pydantic model to write
            path: The file path to save the YAML
        """
        with open(path, "w", encoding="utf-8") as f:
            f.write(cls._yaml_dump(model))

    @staticmethod
    def _read_yaml(path: PathLike) -> dict:
        """
        Reads a YAML file and returns its contents as a dictionary.

        Loads and parses a YAML file into a Python dictionary.

        Args:
            path: The file path to read

        Returns:
            A dictionary representation of the YAML file
        """
        with open(path, "r", encoding="utf-8") as f:
            return yaml.safe_load(f)

    def prompt_input(self) -> bool:
        """
        Prompts the user to confirm whether to generate a manifest.

        Provides user interaction to confirm manifest generation for the
        watchdog service.

        Returns:
            True if the user confirms, False otherwise

        Example:
            ```python
            # Interactive manifest generation:
            settings = WatchdogSettings(
                destination="//server/data",
                project_name="my_project"
            )
            service = WatchdogDataTransferService(source="C:/data", settings=settings)
            if service.prompt_input():
                service.transfer()
                print("Manifest generation confirmed")
            else:
                print("Manifest generation cancelled")
            ```
        """
        return self._ui_helper.prompt_yes_no_question("Would you like to generate a watchdog manifest (Y/N)?")

    @classmethod
    def build_runner(
        cls,
        settings: WatchdogSettings,
        aind_session_data_mapper: Promise[[Launcher], TSessionMapper] | TSessionMapper,
    ) -> Callable[[Launcher], "WatchdogDataTransferService[TSessionMapper]"]:
        """
        A factory method for creating the watchdog service.

        Args:
            settings: The watchdog settings.
            aind_session_data_mapper: The aind session data mapper.

        Returns:
            A factory for WatchdogDataTransferService.
        """

        def _from_launcher(
            launcher: Launcher,
        ) -> "WatchdogDataTransferService":
            """Inner callable to create the service from a launcher"""
            _aind_session_data_mapper = (
                aind_session_data_mapper.result
                if isinstance(aind_session_data_mapper, Promise)
                else aind_session_data_mapper
            )

            if not _aind_session_data_mapper.is_mapped():
                raise ValueError("Data mapper has not mapped yet. Cannot create watchdog.")

            _settings = settings.model_copy()

            _session = launcher.get_session(strict=True)
            _settings.destination = Path(_settings.destination) / _session.subject
            launcher.copy_logs()
            service = cls(
                source=launcher.session_directory,
                settings=_settings,
                session_name=_session.session_name,
            ).with_aind_session_data_mapper(_aind_session_data_mapper)
            service.transfer()
            return service

        return _from_launcher



import subprocess
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

from clabe.apps import BonsaiApp, BonsaiAppSettings, PythonScriptApp


@pytest.fixture
def bonsai_app(mock_ui_helper) -> BonsaiApp:
    """BonsaiApp fixture."""
    workflow = Path("test_workflow.bonsai")
    executable = Path("bonsai/bonsai.exe")
    settings = BonsaiAppSettings(executable=executable, workflow=workflow)
    app = BonsaiApp(settings=settings, ui_helper=mock_ui_helper)
    return app


class TestBonsaiApp:
    """Test BonsaiApp."""

    @patch("clabe.apps._bonsai.run_bonsai_process")
    @patch("pathlib.Path.exists", return_value=True)
    def test_run(self, mock_pathlib: MagicMock, mock_run_bonsai_process: MagicMock, bonsai_app: BonsaiApp) -> None:
        """Test run."""
        mock_result = MagicMock(spec=subprocess.CompletedProcess)
        mock_run_bonsai_process.return_value = mock_result
        result = bonsai_app.run()
        assert result == mock_result
        mock_run_bonsai_process.assert_called_once()

    def test_validate(self, bonsai_app: BonsaiApp) -> None:
        """Test validate."""
        with patch("pathlib.Path.exists", return_value=True):
            assert bonsai_app.validate()

    def test_validate_missing_file(self, bonsai_app: BonsaiApp) -> None:
        """Test validate missing file."""
        with patch("pathlib.Path.exists", side_effect=[False, True, True]):
            with pytest.raises(FileNotFoundError):
                bonsai_app.validate()

    def test_result_property(self, bonsai_app: BonsaiApp) -> None:
        """Test result property."""
        with pytest.raises(RuntimeError):
            _ = bonsai_app.result
        mock_result = MagicMock(spec=subprocess.CompletedProcess)
        bonsai_app._result = mock_result
        assert bonsai_app.result == mock_result

    def test_output_from_result(self, mock_ui_helper, bonsai_app: BonsaiApp) -> None:
        """Test output from result."""
        mock_ui_helper._prompt_yes_no_question.return_value = True
        mock_result = MagicMock(spec=subprocess.CompletedProcess)
        mock_result.stdout = "output"
        mock_result.stderr = ""
        bonsai_app._result = mock_result
        with patch.object(mock_result, "check_returncode", side_effect=subprocess.CalledProcessError(1, "cmd")):
            with pytest.raises(subprocess.CalledProcessError):
                bonsai_app.output_from_result(allow_stderr=True)
        with patch.object(mock_result, "check_returncode", return_value=None):
            assert bonsai_app.output_from_result(allow_stderr=True) == bonsai_app

    def test_prompt_visualizer_layout_input(self, mock_ui_helper, bonsai_app: BonsaiApp) -> None:
        """Test prompt visualizer layout input."""
        mock_ui_helper._prompt_pick_from_list.return_value = "picked_layout.bonsai.layout"
        with patch("glob.glob", return_value=["layout1.bonsai.layout", "layout2.bonsai.layout"]):
            layout = bonsai_app.prompt_visualizer_layout_input()
            assert str(layout) == "picked_layout.bonsai.layout"


@pytest.fixture
def python_script_app() -> PythonScriptApp:
    """PythonScriptApp fixture."""
    app = PythonScriptApp(
        script="test_script.py",
        project_directory=Path("/test/project").as_posix(),
        optional_toml_dependencies=["dep1", "dep2"],
        append_python_exe=True,
        timeout=30,
    )
    return app


class TestPythonScriptApp:
    """Test PythonScriptApp."""

    @patch("subprocess.run")
    def test_create_environment(self, mock_run: MagicMock, python_script_app: PythonScriptApp) -> None:
        """Test create environment."""
        mock_run.return_value = MagicMock(returncode=0)
        result = python_script_app.create_environment()
        mock_run.assert_called_once()
        assert result.returncode == 0

    @patch("subprocess.run")
    @patch("clabe.apps._python_script.PythonScriptApp._has_venv", return_value=True)
    def test_run(self, mock_has_env: MagicMock, mock_run: MagicMock, python_script_app: PythonScriptApp) -> None:
        """Test run."""
        mock_run.return_value = MagicMock(returncode=0)
        result = python_script_app.run()
        mock_run.assert_called_once()
        assert result.returncode == 0

    def test_output_from_result_success(self, python_script_app: PythonScriptApp) -> None:
        """Test output from result success."""
        python_script_app._result = subprocess.CompletedProcess(args="test", returncode=0, stdout="output", stderr="")
        result = python_script_app.output_from_result()
        assert result == python_script_app

    def test_output_from_result_failure(self, python_script_app: PythonScriptApp) -> None:
        """Test output from result failure."""
        python_script_app._result = subprocess.CompletedProcess(
            args="test", returncode=1, stdout="output", stderr="error"
        )
        with pytest.raises(subprocess.CalledProcessError):
            python_script_app.output_from_result()

    def test_result_property(self, python_script_app: PythonScriptApp) -> None:
        """Test result property."""
        with pytest.raises(RuntimeError):
            _ = python_script_app.result

    def test_add_uv_project_directory(self, python_script_app: PythonScriptApp) -> None:
        """Test add uv project directory."""
        assert python_script_app._add_uv_project_directory() == f" --directory {Path('/test/project').resolve()}"

    def test_add_uv_optional_toml_dependencies(self, python_script_app: PythonScriptApp) -> None:
        """Test add uv optional toml dependencies."""
        assert python_script_app._add_uv_optional_toml_dependencies() == "--extra dep1 --extra dep2"



import os
from pathlib import Path
from typing import Optional, Tuple, Type

from pydantic import Field
from pydantic_settings import (
    BaseSettings,
    CliImplicitFlag,
    PydanticBaseSettingsSource,
    YamlConfigSettingsSource,
)

from ..services import ServiceSettings


class LauncherCliArgs(ServiceSettings, cli_prog_name="clabe", cli_kebab_case=True):
    """
    Base class for CLI arguments using Pydantic for validation and configuration.

    Attributes:
        data_dir (os.PathLike): The data directory where to save the data.
        repository_dir (Optional[os.PathLike]): The repository root directory.
        create_directories (CliImplicitFlag[bool]): Whether to create necessary directory structure.
        debug_mode (CliImplicitFlag[bool]): Whether to run in debug mode.
        allow_dirty (CliImplicitFlag[bool]): Whether to allow running with a dirty repository.
        skip_hardware_validation (CliImplicitFlag[bool]): Whether to skip hardware validation.
        subject (Optional[str]): The name of the subject. If None, will be prompted later.
        task_logic_path (Optional[os.PathLike]): Path to the task logic schema. If None, will be prompted later.
        rig_path (Optional[os.PathLike]): Path to the rig schema. If None, will be prompted later.
        temp_dir (os.PathLike): Directory used for launcher temp files.

    Example:
        # Create CLI args from command line
        args = LauncherCliArgs()

        # Create with specific values
        args = LauncherCliArgs(
            data_dir="/path/to/data",
            debug_mode=True,
            subject="mouse_001"
        )

        # Access properties
        print(f"Data directory: {args.data_dir}")
        print(f"Debug mode: {args.debug_mode}")
    """

    data_dir: os.PathLike = Field(description="The data directory where to save the data")
    repository_dir: Optional[os.PathLike] = Field(default=None, description="The repository root directory")
    create_directories: CliImplicitFlag[bool] = Field(
        default=False, description="Whether to create directory structure necessary for the launcher"
    )
    debug_mode: CliImplicitFlag[bool] = Field(default=False, description="Whether to run in debug mode")
    allow_dirty: CliImplicitFlag[bool] = Field(
        default=False, description="Whether to allow the launcher to run with a dirty repository"
    )
    skip_hardware_validation: CliImplicitFlag[bool] = Field(
        default=False, description="Whether to skip hardware validation"
    )
    subject: Optional[str] = Field(default=None, description="The name of the subject. If None, will be prompted later")
    temp_dir: os.PathLike = Field(
        default=Path("local/.temp"), description="The directory used for the launcher temp files"
    )

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> Tuple[PydanticBaseSettingsSource, ...]:
        """
        Customizes the order of settings sources for the CLI.

        Defines the priority order for configuration sources, with initialization settings
        taking precedence, followed by YAML files, environment variables, dotenv files,
        and finally file secrets.

        Args:
            settings_cls: The settings class
            init_settings: Initial settings source
            env_settings: Environment variable settings source
            dotenv_settings: Dotenv settings source
            file_secret_settings: File secret settings source

        Returns:
            Tuple[PydanticBaseSettingsSource, ...]: Ordered tuple of settings sources

        Example:
            # This method is automatically called by Pydantic
            # when creating a LauncherCliArgs instance. Settings are loaded
            # in this priority order:
            # 1. init_settings (constructor arguments)
            # 2. YAML config files
            # 3. Environment variables
            # 4. .env files
            # 5. File secrets
            args = LauncherCliArgs(data_dir="/override/path")  # init_settings
        """
        return (
            init_settings,
            YamlConfigSettingsSource(settings_cls),
            env_settings,
            dotenv_settings,
            file_secret_settings,
        )



from __future__ import annotations

import glob
import logging
import os
from pathlib import Path
from typing import Callable, ClassVar, Generic, List, Optional, Self

import pydantic
from aind_behavior_curriculum import TrainerState
from aind_behavior_services.utils import model_from_json_file
from typing_extensions import override

from clabe import launcher

from .. import ui
from ..services import ServiceSettings
from ..utils import ByAnimalFiles
from ..utils.aind_auth import validate_aind_username
from ._base import TRig, TSession, TTaskLogic

logger = logging.getLogger(__name__)


class DefaultBehaviorPickerSettings(ServiceSettings):
    """
    Settings for the default behavior picker.

    Attributes:
        config_library_dir: The directory where configuration files are stored.
    """

    __yml_section__: ClassVar[Optional[str]] = "default_behavior_picker"

    config_library_dir: os.PathLike


class DefaultBehaviorPicker(Generic[TRig, TSession, TTaskLogic]):
    """
    A picker class for selecting rig, session, and task logic configurations for behavior experiments.

    This class provides methods to initialize directories, pick configurations, and prompt user inputs
    for various components of the experiment setup. It manages the configuration library structure
    and user interactions for selecting experiment parameters.

    Attributes:
        RIG_SUFFIX (str): Directory suffix for rig configurations
        SUBJECT_SUFFIX (str): Directory suffix for subject configurations
        TASK_LOGIC_SUFFIX (str): Directory suffix for task logic configurations

    Example:
        ```python
        # Create settings for the picker
        settings = DefaultBehaviorPickerSettings(config_library_dir="config_dir")

        # Create a default behavior picker
        picker = DefaultBehaviorPicker(
            launcher=some_launcher_instance,
            settings=settings,
        )
        # Initialize and pick configurations
        picker.initialize()
        rig = picker.pick_rig()
        session = picker.pick_session()
        task_logic = picker.pick_task_logic()
        ```
    """

    RIG_SUFFIX: str = "Rig"
    SUBJECT_SUFFIX: str = "Subjects"
    TASK_LOGIC_SUFFIX: str = "TaskLogic"

    @override
    def __init__(
        self,
        *,
        settings: DefaultBehaviorPickerSettings,
        ui_helper: Optional[ui.UiHelper] = None,
        experimenter_validator: Optional[Callable[[str], bool]] = validate_aind_username,
        **kwargs,
    ):
        """
        Initializes the DefaultBehaviorPicker.

        Args:
            settings: Settings containing configuration including config_library_dir
            ui_helper: Helper for user interface interactions
            experimenter_validator: Function to validate the experimenter's username. If None, no validation is performed
            **kwargs: Additional keyword arguments
        """
        self._ui_helper = ui_helper
        self._launcher: launcher.Launcher[TRig, TSession, TTaskLogic]
        self._settings = settings
        self._experimenter_validator = experimenter_validator
        self._trainer_state: Optional[TrainerState] = None

    def register_ui_helper(self, ui_helper: ui.UiHelper) -> Self:
        """
        Registers a UI helper with the picker.

        Associates a UI helper instance with this picker for user interactions.

        Args:
            ui_helper: The UI helper to register

        Returns:
            Self: The picker instance for method chaining

        Raises:
            ValueError: If a UI helper is already registered
        """
        if self._ui_helper is None:
            self._ui_helper = ui_helper
        else:
            raise ValueError("UI Helper is already registered")
        return self

    @property
    def has_ui_helper(self) -> bool:
        """
        Checks if a UI helper is registered.

        Returns:
            bool: True if a UI helper is registered, False otherwise
        """
        return self._ui_helper is not None

    @property
    def ui_helper(self) -> ui.UiHelper:
        """
        Retrieves the registered UI helper.

        Returns:
            DefaultUIHelper: The registered UI helper

        Raises:
            ValueError: If no UI helper is registered
        """
        if self._ui_helper is None:
            raise ValueError("UI Helper is not registered")
        return self._ui_helper

    @property
    def trainer_state(self) -> TrainerState:
        """
        Returns the current trainer state.

        Returns:
            TrainerState: The current trainer state.

        Raises:
            ValueError: If the trainer state is not set.
        """
        if self._trainer_state is None:
            raise ValueError("Trainer state not set.")
        return self._trainer_state

    @property
    def config_library_dir(self) -> Path:
        """
        Returns the path to the configuration library directory.

        Returns:
            Path: The configuration library directory.
        """
        return Path(self._settings.config_library_dir)

    @property
    def rig_dir(self) -> Path:
        """
        Returns the path to the rig configuration directory.

        Returns:
            Path: The rig configuration directory.
        """
        if self._launcher is None:
            raise ValueError("Launcher is not initialized. Call initialize(launcher) first.")
        return Path(os.path.join(self.config_library_dir, self.RIG_SUFFIX, self._launcher.computer_name))

    @property
    def subject_dir(self) -> Path:
        """
        Returns the path to the subject configuration directory.

        Returns:
            Path: The subject configuration directory.
        """
        return Path(os.path.join(self.config_library_dir, self.SUBJECT_SUFFIX))

    @property
    def task_logic_dir(self) -> Path:
        """
        Returns the path to the task logic configuration directory.

        Returns:
            Path: The task logic configuration directory.
        """
        return Path(os.path.join(self.config_library_dir, self.TASK_LOGIC_SUFFIX))

    def initialize(self, launcher: launcher.Launcher[TRig, TSession, TTaskLogic]) -> None:
        """
        Initializes the picker by creating required directories if needed.
        """
        self._launcher = launcher
        self._ui_helper = launcher.ui_helper
        if self._launcher.settings.create_directories:
            self._create_directories(launcher)

    def _create_directories(self, launcher: launcher.Launcher[TRig, TSession, TTaskLogic]) -> None:
        """
        Creates the required directories for configuration files.

        Creates the configuration library directory and all required subdirectories
        for storing rig, task logic, and subject configurations.
        """
        launcher.create_directory(self.config_library_dir)
        launcher.create_directory(self.task_logic_dir)
        launcher.create_directory(self.rig_dir)
        launcher.create_directory(self.subject_dir)

    def pick_rig(self, launcher: launcher.Launcher[TRig, TSession, TTaskLogic]) -> TRig:
        """
        Prompts the user to select a rig configuration file.

        Searches for available rig configuration files and either automatically
        selects a single file or prompts the user to choose from multiple options.

        Returns:
            TRig: The selected rig configuration.

        Raises:
            ValueError: If no rig configuration files are found or an invalid choice is made.
        """
        rig = launcher.get_rig()
        if rig is not None:
            logger.info("Rig already set in launcher. Using existing rig.")
            return rig
        available_rigs = glob.glob(os.path.join(self.rig_dir, "*.json"))
        if len(available_rigs) == 0:
            logger.error("No rig config files found.")
            raise ValueError("No rig config files found.")
        elif len(available_rigs) == 1:
            logger.info("Found a single rig config file. Using %s.", {available_rigs[0]})
            rig = model_from_json_file(available_rigs[0], launcher.get_rig_model())
            launcher.set_rig(rig)
            return rig
        else:
            while True:
                try:
                    path = self.ui_helper.prompt_pick_from_list(available_rigs, prompt="Choose a rig:")
                    if not isinstance(path, str):
                        raise ValueError("Invalid choice.")
                    rig = model_from_json_file(path, launcher.get_rig_model())
                    logger.info("Using %s.", path)
                    launcher.set_rig(rig)
                    return rig
                except pydantic.ValidationError as e:
                    logger.error("Failed to validate pydantic model. Try again. %s", e)
                except ValueError as e:
                    logger.info("Invalid choice. Try again. %s", e)

    def pick_session(self, launcher: launcher.Launcher[TRig, TSession, TTaskLogic]) -> TSession:
        """
        Prompts the user to select or create a session configuration.

        Collects experimenter information, subject selection, and session notes
        to create a new session configuration with appropriate metadata.

        Returns:
            TSession: The created or selected session configuration.
        """
        if (session := launcher.get_session()) is not None:
            logger.info("Session already set in launcher. Using existing session.")
            return session

        experimenter = self.prompt_experimenter(strict=True)
        if launcher.subject is not None:
            logger.info("Subject provided via CLABE: %s", launcher.subject)
            subject = launcher.subject
        else:
            subject = self.choose_subject(self.subject_dir)
            launcher.subject = subject

            if not (self.subject_dir / subject).exists():
                logger.info("Directory for subject %s does not exist. Creating a new one.", subject)
                os.makedirs(self.subject_dir / subject)

        notes = self.ui_helper.prompt_text("Enter notes: ")
        session = launcher.get_session_model()(
            experiment="",  # Will be set later
            root_path=str(Path(launcher.settings.data_dir).resolve() / subject),
            subject=subject,
            notes=notes,
            experimenter=experimenter if experimenter is not None else [],
            commit_hash=launcher.repository.head.commit.hexsha,
            allow_dirty_repo=launcher.settings.debug_mode or launcher.settings.allow_dirty,
            skip_hardware_validation=launcher.settings.skip_hardware_validation,
            experiment_version="",  # Will be set later
        )
        launcher.set_session(session)
        return session

    def pick_task_logic(self, launcher: launcher.Launcher[TRig, TSession, TTaskLogic]) -> TTaskLogic:
        """
        Prompts the user to select or create a task logic configuration.

        Attempts to load task logic in the following order:
        1. From CLI if already set
        2. From subject-specific folder
        3. From user selection in task logic library

        Returns:
            TTaskLogic: The created or selected task logic configuration.

        Raises:
            ValueError: If no valid task logic file is found.
        """
        if (task_logic := launcher.get_task_logic()) is not None:
            logger.info("Task logic already set in launcher. Using existing task logic.")
            self._sync_session_metadata(launcher)
            return task_logic

        # Else, we check inside the subject folder for an existing task file
        try:
            if launcher.subject is None:
                logger.error("No subject set in launcher. Cannot load task logic.")
                raise ValueError("No subject set in launcher.")
            f = self.subject_dir / launcher.subject / (ByAnimalFiles.TASK_LOGIC.value + ".json")
            logger.info("Attempting to load task logic from subject folder: %s", f)
            task_logic = model_from_json_file(f, launcher.get_task_logic_model())
        except (ValueError, FileNotFoundError, pydantic.ValidationError) as e:
            logger.warning("Failed to find a valid task logic file. %s", e)
        else:
            logger.info("Found a valid task logic file in subject folder!")
            _is_manual = not self.ui_helper.prompt_yes_no_question("Would you like to use this task logic?")
            if not _is_manual:
                if task_logic is not None:
                    launcher.set_task_logic(task_logic)
                    return task_logic
                else:
                    logger.error("No valid task logic file found in subject folder.")
                    raise ValueError("No valid task logic file found.")
            else:
                task_logic = None

        # If not found, we prompt the user to choose/enter a task logic file
        while task_logic is None:
            try:
                _path = Path(os.path.join(self.config_library_dir, self.task_logic_dir))
                available_files = glob.glob(os.path.join(_path, "*.json"))
                if len(available_files) == 0:
                    break
                path = self.ui_helper.prompt_pick_from_list(available_files, prompt="Choose a task logic:")
                if not isinstance(path, str):
                    raise ValueError("Invalid choice.")
                if not os.path.isfile(path):
                    raise FileNotFoundError(f"File not found: {path}")
                task_logic = model_from_json_file(path, launcher.get_task_logic_model())
                logger.info("User entered: %s.", path)
            except pydantic.ValidationError as e:
                logger.error("Failed to validate pydantic model. Try again. %s", e)
            except (ValueError, FileNotFoundError) as e:
                logger.info("Invalid choice. Try again. %s", e)
        if task_logic is None:
            logger.error("No task logic file found.")
            raise ValueError("No task logic file found.")

        launcher.set_task_logic(task_logic)
        self._sync_session_metadata(launcher)
        return task_logic

    def _sync_session_metadata(self, launcher: launcher.Launcher[TRig, TSession, TTaskLogic]) -> None:
        """Syncs metadata across session, task_logic and rig models"""
        task_logic = launcher.get_task_logic(strict=True)
        session = launcher.get_session(strict=True)
        session.experiment = task_logic.name
        session.experiment_version = task_logic.version

    def pick_trainer_state(self, launcher: launcher.Launcher[TRig, TSession, TTaskLogic]) -> TrainerState:
        """
        Prompts the user to select or create a trainer state configuration.

        Attempts to load trainer state in the following order:
        1. If task_logic already exists in launcher, will return an empty TrainerState
        2. From subject-specific folder

        It will launcher.set_task_logic if the deserialized TrainerState is valid.

        Returns:
            TrainerState: The deserialized TrainerState object.

        Raises:
            ValueError: If no valid task logic file is found.
        """
        if (launcher.get_task_logic()) is not None:
            logger.info("Task logic already set in launcher. Cannot inject a trainer state.")
            self._trainer_state = TrainerState(curriculum=None, stage=None, is_on_curriculum=False)
        else:
            try:
                if launcher.subject is None:
                    logger.error("No subject set in launcher. Cannot load trainer state.")
                    raise ValueError("No subject set in launcher.")
                f = self.subject_dir / launcher.subject / (ByAnimalFiles.TRAINER_STATE.value + ".json")
                logger.info("Attempting to load trainer state from subject folder: %s", f)
                trainer_state = model_from_json_file(f, TrainerState)
                if trainer_state.stage is None:
                    raise ValueError("Trainer state stage is None, cannot use this trainer state.")
            except (ValueError, FileNotFoundError, pydantic.ValidationError) as e:
                logger.error("Failed to find a valid task logic file. %s", e)
                raise e
            else:
                self._trainer_state = trainer_state
                launcher.set_task_logic(trainer_state.stage.task)

        if not self._trainer_state.is_on_curriculum:
            logging.warning("Deserialized TrainerState is NOT on curriculum.")

        self._sync_session_metadata(launcher)
        return self.trainer_state

    def choose_subject(self, directory: str | os.PathLike) -> str:
        """
        Prompts the user to select or manually enter a subject name.

        Allows the user to either type a new subject name or select from
        existing subject directories.

        Args:
            directory: Path to the directory containing subject folders

        Returns:
            str: The selected or entered subject name.

        Example:
            ```python
            # Choose a subject from the subjects directory
            subject = picker.choose_subject("Subjects")
            ```
        """
        subject = None
        while subject is None:
            subject = self.ui_helper.input("Enter subject name: ")
            if subject == "":
                subject = self.ui_helper.prompt_pick_from_list(
                    [
                        os.path.basename(folder)
                        for folder in os.listdir(directory)
                        if os.path.isdir(os.path.join(directory, folder))
                    ],
                    prompt="Choose a subject:",
                    allow_0_as_none=True,
                )
            else:
                return subject

        return subject

    def prompt_experimenter(self, strict: bool = True) -> Optional[List[str]]:
        """
        Prompts the user to enter the experimenter's name(s).

        Accepts multiple experimenter names separated by commas or spaces.
        Validates names using the configured validator function if provided.

        Args:
            strict: Whether to enforce non-empty input

        Returns:
            Optional[List[str]]: List of experimenter names.

        Example:
            ```python
            # Prompt for experimenter with validation
            names = picker.prompt_experimenter(strict=True)
            print("Experimenters:", names)
            ```
        """
        experimenter: Optional[List[str]] = None
        while experimenter is None:
            _user_input = self.ui_helper.prompt_text("Experimenter name: ")
            experimenter = _user_input.replace(",", " ").split()
            if strict & (len(experimenter) == 0):
                logger.info("Experimenter name is not valid. Try again.")
                experimenter = None
            else:
                if self._experimenter_validator:
                    for name in experimenter:
                        if not self._experimenter_validator(name):
                            logger.warning("Experimenter name: %s, is not valid. Try again", name)
                            experimenter = None
                            break
        return experimenter



import datetime
import logging
import os
from pathlib import Path
from typing import TYPE_CHECKING, TypeVar

import aind_behavior_services.utils as utils
import rich.logging
import rich.style

if TYPE_CHECKING:
    from ..launcher import Launcher

    TLauncher = TypeVar("TLauncher", bound="Launcher")
else:
    TLauncher = TypeVar("TLauncher")

TLogger = TypeVar("TLogger", bound=logging.Logger)

log_fmt = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
datetime_fmt = "%Y-%m-%dT%H%M%S%z"


class _SeverityHighlightingHandler(rich.logging.RichHandler):
    """
    A custom logging handler that highlights log messages based on severity.

    This handler extends RichHandler to provide visual highlighting for error and critical
    log messages using different styles and colors for better visibility.

    Attributes:
        error_style (rich.style.Style): Style for error level messages
        critical_style (rich.style.Style): Style for critical level messages
    """

    def __init__(self, *args, **kwargs):
        """
        Initializes the severity highlighting handler.

        Args:
            *args: Arguments passed to the parent RichHandler
            **kwargs: Keyword arguments passed to the parent RichHandler (highlighter is removed)
        """
        # I don't think this is necessary, but just in case, better to fail early
        if "highlighter" in kwargs:
            del kwargs["highlighter"]
        super().__init__(*args, **kwargs)

        self.error_style = rich.style.Style(color="white", bgcolor="red")
        self.critical_style = rich.style.Style(color="white", bgcolor="red", bold=True)

    def render_message(self, record, message):  # type: ignore[override]
        """
        Renders log messages with severity-based styling.

        Applies different visual styles to log messages based on their severity level,
        with special formatting for error and critical messages.

        Args:
            record: The log record containing message metadata
            message: The log message to render

        Returns:
            str: The styled message string
        """
        if record.levelno >= logging.CRITICAL:
            return f"[{self.critical_style}]{message}[/]"
        elif record.levelno >= logging.ERROR:
            return f"[{self.error_style}]{message}[/]"
        else:
            return message


rich_handler = _SeverityHighlightingHandler(rich_tracebacks=True, show_time=False)


class _TzFormatter(logging.Formatter):
    """
    A custom logging formatter that supports timezone-aware timestamps.

    This formatter extends the standard logging.Formatter to provide timezone-aware
    timestamp formatting for log records.

    Attributes:
        _tz (Optional[timezone]): The timezone to use for formatting timestamps
    """

    def __init__(self, *args, **kwargs):
        """
        Initializes the formatter with optional timezone information.

        Args:
            *args: Positional arguments for the base Formatter class
            **kwargs: Keyword arguments for the base Formatter class.
                      The 'tz' keyword can be used to specify a timezone
        """
        self._tz = kwargs.pop("tz", None)
        super().__init__(*args, **kwargs)

    def formatTime(self, record, datefmt=None) -> str:
        """
        Formats the time of a log record using the specified timezone.

        Converts the log record timestamp to the configured timezone and formats
        it using the AIND behavior services datetime formatting utilities.

        Args:
            record: The log record to format
            datefmt: An optional date format string (unused)

        Returns:
            str: A string representation of the formatted time
        """
        record_time = datetime.datetime.fromtimestamp(record.created, tz=self._tz)
        return utils.format_datetime(record_time)


utc_formatter = _TzFormatter(log_fmt, tz=datetime.timezone.utc)


def add_file_handler(logger: TLogger, output_path: os.PathLike) -> TLogger:
    """
    Adds a file handler to the logger to write logs to a file.

    Creates a new file handler with UTC timezone formatting and adds it to the
    specified logger for persistent log storage.

    Args:
        logger: The logger to which the file handler will be added
        output_path: The path to the log file

    Returns:
        TLogger: The logger with the added file handler
    """
    file_handler = logging.FileHandler(Path(output_path), encoding="utf-8", mode="w")
    file_handler.setFormatter(utc_formatter)
    logger.addHandler(file_handler)
    return logger


def shutdown_logger(logger: TLogger) -> TLogger:
    """
    Shuts down the logger by closing all file handlers and calling logging.shutdown().

    Performs a complete shutdown of the logging system, ensuring all file handlers
    are properly closed and resources are released.

    Args:
        logger: The logger to shut down

    Returns:
        TLogger: The logger with closed file handlers
    """
    close_file_handlers(logger)
    logging.shutdown()
    return logger


def close_file_handlers(logger: TLogger) -> TLogger:
    """
    Closes all file handlers associated with the logger.

    Iterates through all handlers associated with the logger and closes any
    file handlers to ensure proper resource cleanup.

    Args:
        logger: The logger whose file handlers will be closed

    Returns:
        TLogger: The logger with closed file handlers
    """
    for handler in logger.handlers:
        if isinstance(handler, logging.FileHandler):
            handler.close()
    return logger



import logging
import shutil
from typing import List, Self

from git import Repo

from .. import ui

logger = logging.getLogger(__name__)

_HAS_GIT = shutil.which("git") is not None


class GitRepository(Repo):
    """
    A wrapper around the `git.Repo` class that provides additional methods
    for managing Git repositories and their submodules.

    This class extends the functionality of GitPython's Repo class with additional
    utilities for repository management, submodule handling, and cleanup operations.
    """

    def __init__(self, *args, **kwargs):
        """
        Initializes the GitRepository instance and validates the presence of Git.

        Args:
            *args: Arguments passed to the parent Repo class
            **kwargs: Keyword arguments passed to the parent Repo class

        Example:
            ```python
            # Initialize with current directory
            repo = GitRepository()

            # Initialize with specific path
            repo = GitRepository(path="/path/to/repo")
            ```
        """
        super().__init__(*args, **kwargs)
        self._validate_git()

    def reset_repo(self) -> Self:
        """
        Resets the repository to the last committed state.

        Performs a hard reset to discard all uncommitted changes in the working directory
        and staging area, reverting to the last commit.

        Returns:
            Self: The current instance for method chaining.

        Example:
            ```python
            repo = GitRepository("/path/to/repo")
            repo.reset_repo()  # Discards all uncommitted changes
            ```
        """
        self.git.reset("--hard")
        return self

    def clean_repo(self) -> Self:
        """
        Cleans the repository by removing untracked files and directories.

        Removes all untracked files and directories from the working tree that are
        not ignored by .gitignore rules.

        Returns:
            Self: The current instance for method chaining.

        Example:
            ```python
            repo = GitRepository("/path/to/repo")
            repo.clean_repo()  # Removes all untracked files and directories
            ```
        """
        self.git.clean("-fd")
        return self

    def is_dirty_with_submodules(self) -> bool:
        """
        Checks if the repository or any of its submodules is dirty.

        A repository is considered dirty if it has uncommitted changes, including
        untracked files. This method also checks all submodules.

        Returns:
            bool: True if the repository or any submodules have uncommitted changes.

        Example:
            ```python
            repo = GitRepository("/path/to/repo")
            if repo.is_dirty_with_submodules():
                print("Repository or submodules have uncommitted changes")
            ```
        """
        _is_dirty_repo = self.is_dirty(untracked_files=True)
        if _is_dirty_repo:
            return True
        return any([submodule.repo.is_dirty(untracked_files=True) for submodule in self.submodules])

    @staticmethod
    def _get_changes(repo: Repo) -> List[str]:
        """
        Gets a list of changed files in the repository.

        Args:
            repo: The Git repository to check for changes

        Returns:
            List[str]: List of file paths that have been modified
        """
        return [item.a_path for item in (repo.index.diff(None) + repo.index.diff("HEAD")) if item.a_path]

    def uncommitted_changes(self) -> List[str]:
        """
        Retrieves a list of unstaged and untracked files in the repository and its submodules.

        Combines modified files, staged changes, and untracked files from both the main
        repository and all submodules.

        Returns:
            List[str]: A list of file paths with uncommitted changes.

        Example:
            ```python
            repo = GitRepository("/path/to/repo")
            changes = repo.uncommitted_changes()
            if changes:
                print(f"Uncommitted changes found: {changes}")
            ```
        """
        untracked_files = self.untracked_files
        changes = self._get_changes(self)
        for submodule in self.submodules:
            changes.extend(self._get_changes(submodule.repo))
            untracked_files.extend(submodule.repo.untracked_files)
        return list(set(changes + untracked_files))

    def force_update_submodules(self) -> Self:
        """
        Updates all submodules to their latest state.

        Forces an update of all Git submodules to match the commit specified
        in the parent repository.

        Returns:
            Self: The current instance for method chaining.

        Example:
            ```python
            repo = GitRepository("/path/to/repo")
            repo.force_update_submodules()  # Updates all submodules
            ```
        """
        self.submodule_update()
        return self

    def submodules_sync(self) -> Self:
        """
        Synchronizes submodule URLs with the parent repository.

        Updates the submodule URLs to match those defined in the parent repository's
        .gitmodules file.

        Returns:
            Self: The current instance for method chaining.

        Example:
            ```python
            repo = GitRepository("/path/to/repo")
            repo.submodules_sync()  # Synchronizes submodule URLs
            ```
        """
        self.git.submodule("sync", "--recursive")
        return self

    def full_reset(self) -> Self:
        """
        Performs a full reset of the repository and its submodules.

        Executes a complete cleanup including resetting the repository, synchronizing
        submodules, updating them, and cleaning untracked files. Also recursively
        resets all submodules.

        Returns:
            Self: The current instance for method chaining.

        Example:
            ```python
            repo = GitRepository("/path/to/repo")
            repo.full_reset()  # Complete cleanup of repo and submodules
            ```
        """
        self.reset_repo().submodules_sync().force_update_submodules().clean_repo()
        _ = [GitRepository(str(sub.abspath)).full_reset() for sub in self.submodules]
        return self

    def try_prompt_full_reset(self, ui_helper: ui.UiHelper, force_reset: bool = False) -> Self:
        """
        Prompts the user to perform a full reset if the repository is dirty.

        Checks if the repository has uncommitted changes and either prompts the user
        or automatically performs a full reset based on the force_reset parameter.

        Args:
            ui_helper: The UI helper for user interaction
            force_reset: Whether to skip the prompt and force a reset

        Returns:
            Self: The current instance for method chaining.

        Example:
            ```python
            repo = GitRepository("/path/to/repo")
            ui_helper = ui.DefaultUIHelper()
            repo.try_prompt_full_reset(ui_helper)  # Prompts user if dirty
            repo.try_prompt_full_reset(ui_helper, force_reset=True)  # Forces reset
            ```
        """
        if force_reset:
            self.full_reset()
            return self
        if self.is_dirty_with_submodules():
            logger.info("Repository is dirty! %s", self.working_dir)
            logger.info("Uncommitted files: %s", self.uncommitted_changes())
            if not force_reset:
                is_reset = ui_helper.prompt_yes_no_question(prompt="Do you want to reset the repository?")
            else:
                is_reset = True
            if is_reset:
                logger.info("Full reset of repository and submodules: %s", self.working_dir)
                self.full_reset()
        return self

    @staticmethod
    def _validate_git() -> bool:
        """
        Validates the presence of the Git executable.

        Checks if Git is installed and available in the system PATH.

        Raises:
            RuntimeError: If Git is not installed or not found in PATH

        Returns:
            bool: True if Git is installed and available
        """
        if not _HAS_GIT:
            logger.error("git executable not detected.")
            raise RuntimeError(
                "git is not installed in this computer. Please install git. https://git-scm.com/downloads"
            )
        return True



import abc
import logging
import typing as t

import pydantic_settings as ps

from .utils import KNOWN_CONFIG_FILES

if t.TYPE_CHECKING:
    from .launcher import Launcher
else:
    Launcher = t.Any

logger = logging.getLogger(__name__)


class Service(abc.ABC):
    """Abstract base class for all services in the application.
    This may be needed in the future to ensure a common interface"""

    ...


class ServiceSettings(ps.BaseSettings, abc.ABC):
    """
    Base class for service settings with YAML configuration support.

    This class provides automatic YAML configuration loading using pydantic-settings. The configuration is loaded from
    files defined in KNOWN_CONFIG_FILES.

    Attributes:
        __yml_section__: Optional class variable to override the config section name

    Example:
        ```python
        # Define a settings class
        class MyServiceSettings(ServiceSettings):
            __yml_section__: ClassVar[str] = "my_service"

            host: str = "localhost"
            port: int = 8080
            enabled: bool = True

        # Usage will automatically load from YAML files
        settings = MyServiceSettings()
        ```
    """

    __yml_section__: t.ClassVar[t.Optional[str]] = None

    @classmethod
    def __init_subclass__(cls, *args, **kwargs):
        """Initializes the subclass and sets up the YAML configuration."""
        super().__init_subclass__(*args, **kwargs)
        cls.model_config.update(
            ps.SettingsConfigDict(yaml_file=KNOWN_CONFIG_FILES, yaml_config_section=cls.__yml_section__, extra="ignore")
        )

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: t.Type[ps.BaseSettings],
        init_settings: ps.PydanticBaseSettingsSource,
        env_settings: ps.PydanticBaseSettingsSource,
        dotenv_settings: ps.PydanticBaseSettingsSource,
        file_secret_settings: ps.PydanticBaseSettingsSource,
    ) -> t.Tuple[ps.PydanticBaseSettingsSource, ...]:
        """
        Customizes the settings sources to include the safe YAML settings source.

        Args:
            settings_cls: The settings class.
            init_settings: The initial settings source.
            env_settings: The environment settings source.
            dotenv_settings: The dotenv settings source.
            file_secret_settings: The file secret settings source.

        Returns:
            A tuple of settings sources.
        """
        return (
            init_settings,
            _SafeYamlSettingsSource(settings_cls),
            env_settings,
            dotenv_settings,
            file_secret_settings,
        )


class _SafeYamlSettingsSource(ps.YamlConfigSettingsSource):
    """
    A safe YAML settings source that does not raise an error if the YAML configuration section is not found.
    """

    def __init__(
        self,
        settings_cls: type[ps.BaseSettings],
        yaml_file: ps.sources.types.PathType | None = ps.sources.types.DEFAULT_PATH,
        yaml_file_encoding: str | None = None,
        yaml_config_section: str | None = None,
    ):
        """
        Initializes the safe YAML settings source.

        Args:
            settings_cls: The settings class.
            yaml_file: The YAML file path.
            yaml_file_encoding: The YAML file encoding.
            yaml_config_section: The YAML configuration section.
        """
        try:
            # pydantic-settings will raise an error if a yaml_config_section is passed but is not found in the yaml file
            # We override this behavior to allow us to have a behavior as if the file did not exist in the first place
            # We may consider raising a more useful error in the future
            super().__init__(settings_cls, yaml_file, yaml_file_encoding, yaml_config_section)
        except KeyError:
            settings_cls.model_config.update({"yaml_config_section": None})
            super().__init__(settings_cls, yaml_file, yaml_file_encoding, None)

    def __call__(self) -> t.Dict[str, t.Any]:
        """
        Calls the settings source and returns the settings dictionary.

        Returns:
            A dictionary of settings.
        """
        try:
            return super().__call__()
        except KeyError:
            return {}



from ._base import Launcher, TModel, TRig, TSession, TTaskLogic
from ._callable_manager import Promise, ignore_errors, run_if
from ._cli import LauncherCliArgs
from ._picker import DefaultBehaviorPicker, DefaultBehaviorPickerSettings

__all__ = [
    "Launcher",
    "TModel",
    "TRig",
    "TSession",
    "TTaskLogic",
    "LauncherCliArgs",
    "DefaultBehaviorPicker",
    "DefaultBehaviorPickerSettings",
    "ignore_errors",
    "run_if",
    "Promise",
]



from __future__ import annotations

import logging
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Callable, List, Optional

from ..services import Service

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from ..launcher import Launcher
else:
    Launcher = Any


class ResourceMonitor(Service):
    """
    A service that monitors and validates resource constraints.

    This service manages a collection of constraints that can be evaluated to ensure
    that system resources meet the requirements for experiment execution.

    Attributes:
        constraints (List[Constraint]): A list of constraints to monitor

    Example:
        ```python
        from clabe.resource_monitor import available_storage_constraint_factory

        # Create monitor with storage constraint
        monitor = ResourceMonitor()
        storage_constraint = available_storage_constraint_factory(
            drive="C:\\", min_bytes=1e9  # 1GB minimum
        )
        monitor.add_constraint(storage_constraint)

        # Validate all constraints
        if monitor.validate():
            print("All constraints satisfied")
        else:
            print("Some constraints failed")
        ```
    """

    def __init__(
        self,
        *args,
        constrains: Optional[List[Constraint]] = None,
        **kwargs,
    ) -> None:
        """
        Initializes the ResourceMonitor.

        Args:
            constrains: A list of constraints to initialize with. Defaults to None
        """
        self.constraints = constrains or []

    def build_runner(self) -> Callable[[Launcher], bool]:
        """
        Builds a runner function that evaluates all constraints.

        Returns:
            A callable that takes a launcher instance and returns True if all constraints are satisfied, False otherwise.
        """

        def _run(launcher: Launcher) -> bool:
            """Inner function to run the resource monitor given a launcher instance."""
            logger.debug("Evaluating resource monitor constraints.")
            if result := not self.evaluate_constraints():
                logger.error("One or more resource monitor constraints failed.")
                raise RuntimeError("Resource monitor constraints failed.")
            return result

        return _run

    def add_constraint(self, constraint: Constraint) -> None:
        """
        Adds a new constraint to the monitor.

        Registers a new constraint for monitoring with this resource monitor.

        Args:
            constraint: The constraint to add

        Example:
            ```python
            monitor = ResourceMonitor()

            # Add storage constraint
            storage_constraint = available_storage_constraint_factory()
            monitor.add_constraint(storage_constraint)

            # Add custom constraint
            def check_memory():
                return psutil.virtual_memory().available > 1e9

            memory_constraint = Constraint(
                name="memory_check",
                constraint=check_memory
            )
            monitor.add_constraint(memory_constraint)
            ```
        """
        self.constraints.append(constraint)

    def remove_constraint(self, constraint: Constraint) -> None:
        """
        Removes a constraint from the monitor.

        Unregisters a previously added constraint from monitoring.

        Args:
            constraint: The constraint to remove

        Example:
            ```python
            monitor = ResourceMonitor()
            monitor.add_constraint(constraint)

            # Later remove it
            monitor.remove_constraint(constraint)
            ```
        """
        self.constraints.remove(constraint)

    def evaluate_constraints(self) -> bool:
        """
        Evaluates all constraints.

        Iterates through all registered constraints and evaluates them, logging
        any failures that occur.

        Returns:
            bool: True if all constraints are satisfied, False otherwise

        Example:
            ```python
            monitor = ResourceMonitor([constraint1, constraint2])

            # Check if all constraints pass
            all_passed = monitor.evaluate_constraints()
            if not all_passed:
                print("Some constraints failed - check logs")
            ```
        """
        for constraint in self.constraints:
            if not constraint():
                logger.error(constraint.on_fail())
                return False
        return True


@dataclass(frozen=True)
class Constraint:
    """
    Represents a resource constraint.

    This class encapsulates a constraint function along with its parameters and
    failure handling logic for resource monitoring.

    Attributes:
        name (str): The name of the constraint
        constraint (Callable[..., bool]): The function to evaluate the constraint
        args (List): Positional arguments for the constraint function
        kwargs (dict): Keyword arguments for the constraint function
        fail_msg_handler (Optional[Callable[..., str]]): A function to generate a failure message

    Example:
        ```python
        # Simple constraint
        def check_disk_space(path, min_gb):
            free_gb = shutil.disk_usage(path).free / (1024**3)
            return free_gb >= min_gb

        constraint = Constraint(
            name="disk_space_check",
            constraint=check_disk_space,
            kwargs={"path": "C:\\", "min_gb": 10},
            fail_msg_handler=lambda path, min_gb: f"Need {min_gb}GB free on {path}"
        )

        simple_constraint = Constraint(
            name="simple_check",
            constraint=lambda x: x > 5)

        # Evaluate constraint
        if constraint():
            print("Constraint passed")
        else:
            print(constraint.on_fail())

        # Evaluate simple constraint
        if simple_constraint():
            print("Simple constraint passed")
        else:
            print(simple_constraint.on_fail())
        ```
    """

    name: str
    constraint: Callable[..., bool]
    args: List = field(default_factory=list)
    kwargs: dict = field(default_factory=dict)
    fail_msg_handler: Optional[Callable[..., str]] = field(default=None)

    def __call__(self) -> bool | Exception:
        """
        Evaluates the constraint.

        Executes the constraint function with the stored arguments and returns
        the result of the evaluation.

        Returns:
            bool | Exception: True if the constraint is satisfied, otherwise raises an exception

        Example:
            ```python
            constraint = Constraint(
                name="test",
                constraint=lambda x: x > 5,
                kwargs={"x": 10}
            )

            result = constraint()  # Returns True
            print(f"Constraint passed: {result}")
            ```
        """
        return self.constraint(*self.args, **self.kwargs)

    def on_fail(self) -> str:
        """
        Generates a failure message if the constraint is not satisfied.

        Uses the registered failure message handler or a default message to
        provide information about constraint failures.

        Returns:
            str: The failure message

        Example:
            ```python
            constraint = Constraint(
                name="memory_check",
                constraint=lambda: False,  # Always fails
                fail_msg_handler=lambda: "Not enough memory available"
            )

            if not constraint():
                print(constraint.on_fail())  # "Not enough memory available"
            ```
        """
        if self.fail_msg_handler:
            return self.fail_msg_handler(*self.args, **self.kwargs)
        else:
            return f"Constraint {self.name} failed."



import platform
from unittest.mock import MagicMock, patch

import pytest

from clabe.utils import aind_auth


@pytest.mark.skipif(platform.system() != "Windows", reason="Windows-only test")
@patch("clabe.utils.aind_auth.ms_active_directory")
@patch("clabe.utils.aind_auth.ldap3")
def test_validate_aind_username_windows_valid(mock_ldap3, mock_ad):
    """Test validate_aind_username on Windows with a valid user."""
    mock_session = MagicMock()
    mock_session.find_user_by_name.return_value = True
    mock_ad.ADDomain.return_value.create_session_as_user.return_value = mock_session

    assert aind_auth.validate_aind_username("testuser")


@pytest.mark.skipif(platform.system() != "Windows", reason="Windows-only test")
@patch("clabe.utils.aind_auth.ms_active_directory")
@patch("clabe.utils.aind_auth.ldap3")
def test_validate_aind_username_windows_invalid(mock_ldap3, mock_ad):
    """Test validate_aind_username on Windows with an invalid user."""
    mock_session = MagicMock()
    mock_session.find_user_by_name.return_value = None
    mock_ad.ADDomain.return_value.create_session_as_user.return_value = mock_session

    assert not aind_auth.validate_aind_username("testuser")


@pytest.mark.skipif(platform.system() != "Windows", reason="Windows-only test")
@patch("clabe.utils.aind_auth.ms_active_directory")
@patch("clabe.utils.aind_auth.ldap3")
def test_validate_aind_username_windows_timeout(mock_ldap3, mock_ad):
    """Test validate_aind_username on Windows with a timeout."""
    with patch("concurrent.futures.ThreadPoolExecutor.submit") as mock_submit:
        mock_submit.side_effect = TimeoutError
        with pytest.raises(TimeoutError):
            aind_auth.validate_aind_username("testuser")



import os
import shutil
from pathlib import Path

from ._base import Constraint


def available_storage_constraint_factory(drive: os.PathLike = Path(r"C:\\"), min_bytes: float = 2e11) -> Constraint:
    """
    Creates a constraint to check if a drive has sufficient available storage.

    This factory function creates a constraint that validates whether the specified
    drive has enough free space to meet the minimum requirements.

    Args:
        drive: The drive to check. Defaults to "C:\\":
        min_bytes: Minimum required free space in bytes. Defaults to 200GB

    Returns:
        Constraint: A constraint object for available storage validation

    Raises:
        ValueError: If the drive path is not valid

    Example:
        ```python
        # Check for default 200GB free space on C: drive
        default_storage_constraint = available_storage_constraint_factory()

        # Check for 1TB free space on D: drive
        large_storage_constraint = available_storage_constraint_factory(
            drive="D:\\",
            min_bytes=1e12  # 1TB
        )

        # Use in resource monitor
        monitor = ResourceMonitor()
        monitor.add_constraint(large_storage_constraint)
        ```
    """
    if not os.path.ismount(drive):
        drive = os.path.splitdrive(drive)[0] + "\\"
    if drive is None:
        raise ValueError("Drive is not valid.")
    return Constraint(
        name="available_storage",
        constraint=lambda drive, min_bytes: shutil.disk_usage(drive).free >= min_bytes,
        args=[],
        kwargs={"drive": drive, "min_bytes": min_bytes},
        fail_msg_handler=lambda drive, min_bytes: f"Drive {drive} does not have enough space.",
    )


def remote_dir_exists_constraint_factory(dir_path: os.PathLike) -> Constraint:
    """
    Creates a constraint to check if a remote directory exists.

    This factory function creates a constraint that validates whether the specified
    directory path exists and is accessible.

    Args:
        dir_path: The path of the directory to check

    Returns:
        Constraint: A constraint object for directory existence validation

    Example:
        ```python
        # Check if network share exists
        network_constraint = remote_dir_exists_constraint_factory(
            "\\\\server\\shared_folder"
        )

        # Check if local directory exists
        local_constraint = remote_dir_exists_constraint_factory(
            "/data/experiments"
        )

        # Use in resource monitor
        monitor = ResourceMonitor()
        monitor.add_constraint(network_constraint)
        monitor.add_constraint(local_constraint)

        if monitor.validate():
            print("All directories accessible")
        ```
    """
    return Constraint(
        name="remote_dir_exists",
        constraint=lambda dir_path: os.path.exists(dir_path),
        args=[],
        kwargs={"dir_path": dir_path},
        fail_msg_handler=lambda dir_path: f"Directory {dir_path} does not exist.",
    )



import os
import subprocess
from datetime import datetime, time
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest
from aind_data_schema.core.session import Session as AdsSession
from aind_data_transfer_service.models.core import Task
from requests.exceptions import HTTPError

from clabe.data_mapper.aind_data_schema import AindDataSchemaSessionDataMapper
from clabe.data_transfer.aind_watchdog import (
    CORE_FILES,
    ManifestConfig,
    WatchConfig,
    WatchdogDataTransferService,
    WatchdogSettings,
)
from clabe.data_transfer.robocopy import RobocopyService, RobocopySettings


@pytest.fixture
def source():
    return Path("source_path")


@pytest.fixture
def ads_session():
    """Mock AdsSession for testing create_manifest_config_from_ads_session method."""
    mock_session = MagicMock(spec=AdsSession)
    mock_session.experimenter_full_name = ["John Doe", "Jane Smith"]
    mock_session.subject_id = "12345"
    mock_session.session_start_time = datetime(2023, 1, 1, 10, 0, 0)

    # Mock data streams with modalities
    mock_modality = MagicMock()
    mock_modality.abbreviation = "behavior"

    mock_data_stream = MagicMock()
    mock_session.data_streams = [mock_data_stream]
    mock_session.data_streams[0].stream_modalities = [mock_modality]

    return mock_session


@pytest.fixture
def aind_data_mapper(ads_session):
    mapper = MagicMock(spec=AindDataSchemaSessionDataMapper)
    mapper.is_mapped.return_value = True
    mapper.mapped = ads_session
    return mapper


@pytest.fixture
def settings():
    return WatchdogSettings(
        destination=Path("destination_path"),
        schedule_time=time(hour=20),
        project_name="test_project",
        platform="behavior",
        capsule_id="capsule_id",
        script={"script_key": ["script_value"]},
        s3_bucket="private",
        mount="mount_path",
        force_cloud_sync=True,
        transfer_endpoint="http://aind-data-transfer-service/api/v2/submit_jobs",
    )


@pytest.fixture
def watchdog_service(mock_ui_helper, source, settings):
    os.environ["WATCHDOG_EXE"] = "watchdog.exe"
    os.environ["WATCHDOG_CONFIG"] = "watchdog_config.yml"

    service = WatchdogDataTransferService(
        source,
        settings=settings,
        validate=False,
        ui_helper=mock_ui_helper,
        session_name="test_session",
    )

    service._manifest_config = ManifestConfig(
        name="test_manifest",
        modalities={"behavior": ["path/to/behavior"], "behavior-videos": ["path/to/behavior-videos"]},
        subject_id=1,
        acquisition_datetime=datetime(2023, 1, 1, 0, 0, 0),
        schemas=["path/to/schema"],
        destination="path/to/destination",
        mount="mount_path",
        processor_full_name="processor_name",
        project_name="test_project",
        schedule_time=settings.schedule_time,
        platform="behavior",
        capsule_id="capsule_id",
        s3_bucket="private",
        script={"script_key": ["script_value"]},
        force_cloud_sync=True,
        transfer_endpoint="http://aind-data-transfer-service/api/v2/submit_jobs",
    )

    service._watch_config = WatchConfig(
        flag_dir="flag_dir",
        manifest_complete="manifest_complete",
    )

    yield service

    # Cleanup
    if "WATCHDOG_EXE" in os.environ:
        del os.environ["WATCHDOG_EXE"]
    if "WATCHDOG_CONFIG" in os.environ:
        del os.environ["WATCHDOG_CONFIG"]


class TestWatchdogDataTransferService:
    def test_initialization(self, watchdog_service, settings):
        assert watchdog_service._settings.destination == settings.destination
        assert watchdog_service._settings.project_name == settings.project_name
        assert watchdog_service._settings.schedule_time == settings.schedule_time
        assert watchdog_service._settings.platform == settings.platform
        assert watchdog_service._settings.capsule_id == settings.capsule_id
        assert watchdog_service._settings.script == settings.script
        assert watchdog_service._settings.s3_bucket == settings.s3_bucket
        assert watchdog_service._settings.mount == settings.mount
        assert watchdog_service._settings.force_cloud_sync == settings.force_cloud_sync
        assert watchdog_service._settings.transfer_endpoint == settings.transfer_endpoint

    @patch("clabe.data_transfer.aind_watchdog.subprocess.check_output")
    def test_is_running(self, mock_check_output, watchdog_service):
        mock_check_output.return_value = (
            "Image Name                     PID Session Name        Session#    Mem Usage\n"
            "========================= ======== ================ =========== ============\n"
            "watchdog.exe                1234 Console                    1    10,000 K\n"
        )
        assert watchdog_service.is_running()

    @patch("clabe.data_transfer.aind_watchdog.subprocess.check_output")
    def test_is_not_running(self, mock_check_output, watchdog_service):
        mock_check_output.return_value = "INFO: No tasks are running which match the specified criteria."
        assert not watchdog_service.is_running()

    @patch("clabe.data_transfer.aind_watchdog.requests.get")
    def test_get_project_names(self, mock_get, watchdog_service):
        mock_response = MagicMock()
        mock_response.ok = True
        mock_response.content = '{"data": ["test_project"]}'
        mock_get.return_value = mock_response
        project_names = watchdog_service._get_project_names()
        assert "test_project" in project_names

    @patch("clabe.data_transfer.aind_watchdog.requests.get")
    def test_get_project_names_fail(self, mock_get, watchdog_service):
        mock_response = MagicMock()
        mock_response.ok = False
        mock_get.return_value = mock_response
        with pytest.raises(Exception):
            watchdog_service._get_project_names()

    @patch(
        "clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.is_running",
        return_value=True,
    )
    @patch(
        "clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.is_valid_project_name",
        return_value=True,
    )
    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService._read_yaml")
    def test_validate_success(self, mock_read_yaml, mock_is_valid_project_name, mock_is_running, watchdog_service):
        mock_read_yaml.return_value = WatchConfig(
            flag_dir="mock_flag_dir", manifest_complete="manifest_complete_dir"
        ).model_dump()
        with patch.object(Path, "exists", return_value=True):
            assert watchdog_service.validate()

    @patch(
        "clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.is_running",
        return_value=False,
    )
    def test_validate_fail(self, mock_is_running, watchdog_service):
        with patch.object(Path, "exists", return_value=False):
            with pytest.raises(FileNotFoundError):
                watchdog_service.validate()

    def test_aind_session_data_mapper_get(self, watchdog_service, aind_data_mapper):
        watchdog_service.with_aind_session_data_mapper(aind_data_mapper)
        assert watchdog_service.aind_session_data_mapper == aind_data_mapper

    def test_aind_session_data_mapper_get_not_set(self, watchdog_service):
        watchdog_service._aind_session_data_mapper = None
        with pytest.raises(ValueError):
            _ = watchdog_service.aind_session_data_mapper

    def test_with_aind_session_data_mapper(self, watchdog_service, aind_data_mapper):
        returned_service = watchdog_service.with_aind_session_data_mapper(aind_data_mapper)
        assert watchdog_service._aind_session_data_mapper == aind_data_mapper
        assert returned_service == watchdog_service

    def test_missing_env_variables(self, source, settings, aind_data_mapper):
        if "WATCHDOG_EXE" in os.environ:
            del os.environ["WATCHDOG_EXE"]
        if "WATCHDOG_CONFIG" in os.environ:
            del os.environ["WATCHDOG_CONFIG"]
        with pytest.raises(ValueError):
            WatchdogDataTransferService(
                source,
                settings=settings,
                validate=False,
            ).with_aind_session_data_mapper(aind_data_mapper)

    @patch("clabe.data_transfer.aind_watchdog.Path.exists", return_value=True)
    def test_find_ads_schemas(self, mock_exists):
        source = "mock_source_path"
        expected_files = [Path(source) / f"{file}.json" for file in CORE_FILES]

        result = WatchdogDataTransferService._find_ads_schemas(Path(source))
        assert result == expected_files

    @patch("clabe.data_transfer.aind_watchdog.Path.mkdir")
    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService._write_yaml")
    def test_dump_manifest_config(self, mock_write_yaml, mock_mkdir, watchdog_service):
        path = Path("flag_dir/manifest_test_manifest.yaml")
        result = watchdog_service.dump_manifest_config()

        assert isinstance(result, Path)
        assert isinstance(path, Path)
        assert result.resolve() == path.resolve()

        mock_write_yaml.assert_called_once()
        mock_mkdir.assert_called_once_with(parents=True, exist_ok=True)

    @patch("clabe.data_transfer.aind_watchdog.Path.mkdir")
    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService._write_yaml")
    def test_dump_manifest_config_custom_path(self, mock_write_yaml, mock_mkdir, watchdog_service):
        custom_path = Path("custom_path/manifest_test_manifest.yaml")
        result = watchdog_service.dump_manifest_config(path=custom_path)

        assert isinstance(result, Path)
        assert isinstance(custom_path, Path)
        assert result.resolve() == custom_path.resolve()
        mock_write_yaml.assert_called_once()
        mock_mkdir.assert_called_once_with(parents=True, exist_ok=True)

    def test_dump_manifest_config_no_manifest_config(self, watchdog_service):
        watchdog_service._manifest_config = None

        with pytest.raises(ValueError):
            watchdog_service.dump_manifest_config()

    def test_dump_manifest_config_no_watch_config(self, watchdog_service):
        watchdog_service._watch_config = None

        with pytest.raises(ValueError):
            watchdog_service.dump_manifest_config()

    def test_from_settings_transfer_args(
        self, watchdog_service: WatchdogDataTransferService, settings: WatchdogSettings
    ):
        settings.upload_tasks = {
            "myTask": Task(job_settings={"input_source": "not_interpolated"}),
            "nestedTask": {"nestedTask": Task(job_settings={"input_source": "not_interpolated_nested"})},
            "myTaskInterpolated": Task(job_settings={"input_source": "interpolated/path/{{ destination }}"}),
            "nestedTaskInterpolated": {
                "nestedTask": Task(job_settings={"input_source": "interpolated/path/{{ destination }}/nested"})
            },
        }
        settings.job_type = "not_default"

        manifest = watchdog_service._manifest_config
        new_watchdog_manifest = watchdog_service._make_transfer_args(
            manifest,
            add_default_tasks=True,
            extra_tasks=settings.upload_tasks or {},
            job_type=settings.job_type,
        )

        transfer_service_args = new_watchdog_manifest.transfer_service_args
        assert transfer_service_args is not None, "Transfer service args are not set"
        tasks = transfer_service_args.upload_jobs[0].tasks
        assert transfer_service_args.upload_jobs[0].job_type == "not_default"
        assert "modality_transformation_settings" in tasks
        assert "gather_preliminary_metadata" in tasks
        assert all(task in tasks for task in ["myTask", "nestedTask", "myTaskInterpolated", "nestedTaskInterpolated"])
        assert (
            Path(tasks["myTaskInterpolated"].job_settings["input_source"]).resolve()
            == Path(f"interpolated/path/{WatchdogDataTransferService._remote_destination_root(manifest)}").resolve()
        )
        assert (
            Path(tasks["nestedTaskInterpolated"]["nestedTask"].job_settings["input_source"]).resolve()
            == Path(
                f"interpolated/path/{WatchdogDataTransferService._remote_destination_root(manifest)}/nested"
            ).resolve()
        )

    def test_make_transfer_args(self, watchdog_service: WatchdogDataTransferService):
        manifest = watchdog_service._manifest_config
        extra_tasks = {
            "myTask": Task(job_settings={"input_source": "not_interpolated"}),
            "nestedTask": {"nestedTask": Task(job_settings={"input_source": "not_interpolated_nested"})},
            "myTaskInterpolated": Task(job_settings={"input_source": "interpolated/path/{{ destination }}"}),
            "nestedTaskInterpolated": {
                "nestedTask": Task(job_settings={"input_source": "interpolated/path/{{ destination }}/nested"})
            },
        }
        assert manifest is not None, "Manifest config is not set"
        new_watchdog_manifest = watchdog_service._make_transfer_args(
            manifest, add_default_tasks=True, extra_tasks=extra_tasks, job_type="not_default"
        )
        transfer_service_args = new_watchdog_manifest.transfer_service_args
        assert transfer_service_args is not None, "Transfer service args are not set"
        tasks = transfer_service_args.upload_jobs[0].tasks
        assert transfer_service_args.upload_jobs[0].job_type == "not_default"
        assert "modality_transformation_settings" in tasks
        assert "gather_preliminary_metadata" in tasks
        assert all(task in tasks for task in ["myTask", "nestedTask", "myTaskInterpolated", "nestedTaskInterpolated"])
        assert (
            Path(tasks["myTaskInterpolated"].job_settings["input_source"]).resolve()
            == Path(f"interpolated/path/{WatchdogDataTransferService._remote_destination_root(manifest)}").resolve()
        )
        assert (
            Path(tasks["nestedTaskInterpolated"]["nestedTask"].job_settings["input_source"]).resolve()
            == Path(
                f"interpolated/path/{WatchdogDataTransferService._remote_destination_root(manifest)}/nested"
            ).resolve()
        )

    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.is_running", return_value=False)
    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.force_restart", return_value=None)
    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.dump_manifest_config")
    def test_transfer_service_not_running_restart_success(
        self,
        mock_dump_manifest_config,
        mock_force_restart,
        mock_is_running,
        watchdog_service,
        aind_data_mapper,
    ):
        mock_is_running.side_effect = [False, True]  # First call returns False, second returns True
        watchdog_service.with_aind_session_data_mapper(aind_data_mapper)
        watchdog_service.transfer()
        mock_force_restart.assert_called_once_with(kill_if_running=False)
        mock_dump_manifest_config.assert_called_once()

    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.is_running", return_value=False)
    @patch(
        "clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.force_restart",
        side_effect=subprocess.CalledProcessError(1, "cmd"),
    )
    def test_transfer_service_not_running_restart_fail(
        self, mock_force_restart, mock_is_running, watchdog_service, aind_data_mapper
    ):
        watchdog_service.with_aind_session_data_mapper(aind_data_mapper)
        with pytest.raises(RuntimeError):
            watchdog_service.transfer()
        mock_force_restart.assert_called_once_with(kill_if_running=False)

    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.is_running", return_value=True)
    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.dump_manifest_config")
    def test_transfer_watch_config_none(
        self,
        mock_dump_manifest_config,
        mock_is_running,
        watchdog_service,
        aind_data_mapper,
    ):
        watchdog_service._watch_config = None
        watchdog_service.with_aind_session_data_mapper(aind_data_mapper)
        with pytest.raises(ValueError):
            watchdog_service.transfer()
        mock_dump_manifest_config.assert_not_called()

    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.is_running", return_value=True)
    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.dump_manifest_config")
    def test_transfer_success(
        self,
        mock_dump_manifest_config,
        mock_is_running,
        watchdog_service,
        aind_data_mapper,
    ):
        watchdog_service.with_aind_session_data_mapper(aind_data_mapper)
        watchdog_service.transfer()
        mock_dump_manifest_config.assert_called_once()

    @patch("clabe.data_transfer.aind_watchdog.Path.exists", return_value=False)
    def test_validate_executable_not_found(self, mock_exists, watchdog_service):
        with pytest.raises(FileNotFoundError):
            watchdog_service.validate()

    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.is_running", return_value=False)
    @patch("clabe.data_transfer.aind_watchdog.Path.exists", return_value=True)
    @patch(
        "clabe.data_transfer.aind_watchdog.WatchdogDataTransferService._read_yaml",
        return_value={"flag_dir": "mock_flag_dir", "manifest_complete": "mock_manifest_complete"},
    )
    def test_validate_service_not_running(self, mock_read_yaml, mock_exists, mock_is_running, watchdog_service):
        assert not watchdog_service.validate()

    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.is_valid_project_name", return_value=False)
    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.is_running", return_value=True)
    @patch("clabe.data_transfer.aind_watchdog.Path.exists", return_value=True)
    @patch(
        "clabe.data_transfer.aind_watchdog.WatchdogDataTransferService._read_yaml",
        return_value={"flag_dir": "mock_flag_dir", "manifest_complete": "mock_manifest_complete"},
    )
    def test_validate_invalid_project_name(
        self,
        mock_read_yaml,
        mock_exists,
        mock_is_running,
        mock_is_valid_project_name,
        watchdog_service,
    ):
        assert not watchdog_service.validate()

    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.is_valid_project_name", side_effect=HTTPError)
    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.is_running", return_value=True)
    @patch("clabe.data_transfer.aind_watchdog.Path.exists", return_value=True)
    @patch(
        "clabe.data_transfer.aind_watchdog.WatchdogDataTransferService._read_yaml",
        return_value={"flag_dir": "mock_flag_dir", "manifest_complete": "mock_manifest_complete"},
    )
    def test_validate_http_error(
        self,
        mock_read_yaml,
        mock_exists,
        mock_is_running,
        mock_is_valid_project_name,
        watchdog_service,
    ):
        with pytest.raises(HTTPError):
            watchdog_service.validate()

    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.is_valid_project_name", return_value=True)
    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService.is_running", return_value=True)
    @patch("clabe.data_transfer.aind_watchdog.Path.exists", return_value=True)
    @patch(
        "clabe.data_transfer.aind_watchdog.WatchdogDataTransferService._read_yaml",
        return_value={"flag_dir": "mock_flag_dir", "manifest_complete": "mock_manifest_complete"},
    )
    def test_validate_success_extended(
        self,
        mock_read_yaml,
        mock_exists,
        mock_is_running,
        mock_is_valid_project_name,
        watchdog_service,
    ):
        assert watchdog_service.validate()

    @patch(
        "clabe.data_transfer.aind_watchdog.WatchdogDataTransferService._get_project_names",
        return_value=["test_project"],
    )
    def test_is_valid_project_name_valid(self, mock_get_project_names, watchdog_service):
        assert watchdog_service.is_valid_project_name()

    @patch(
        "clabe.data_transfer.aind_watchdog.WatchdogDataTransferService._get_project_names",
        return_value=["other_project"],
    )
    def test_is_valid_project_name_invalid(self, mock_get_project_names, watchdog_service):
        assert not watchdog_service.is_valid_project_name()

    @patch(
        "clabe.data_transfer.aind_watchdog.WatchdogDataTransferService._get_project_names",
        return_value=["other_project"],
    )
    @patch("clabe.data_transfer.aind_watchdog.WatchdogDataTransferService._find_ads_schemas", return_value=[])
    def test_create_manifest_config_from_ads_session_invalid_project_name(
        self, mock_find_ads_schemas, mock_get_project_names, watchdog_service, aind_data_mapper
    ):
        watchdog_service._validate_project_name = True
        with pytest.raises(ValueError):
            watchdog_service._create_manifest_config_from_ads_session(aind_data_mapper.mapped)


@pytest.fixture
def robocopy_source():
    return Path("source_path")


@pytest.fixture
def robocopy_settings():
    return RobocopySettings(
        destination=Path("destination_path"),
        log=Path("log_path"),
        extra_args="/MIR",
        delete_src=True,
        overwrite=True,
        force_dir=False,
    )


@pytest.fixture
def robocopy_service(mock_ui_helper, robocopy_source, robocopy_settings):
    return RobocopyService(
        source=robocopy_source,
        settings=robocopy_settings,
        ui_helper=mock_ui_helper,
    )


class TestRobocopyService:
    def test_initialization(self, robocopy_service, robocopy_source, robocopy_settings):
        assert robocopy_service.source == robocopy_source
        assert robocopy_service._settings.destination == robocopy_settings.destination
        assert robocopy_service._settings.log == robocopy_settings.log
        assert robocopy_service._settings.extra_args == robocopy_settings.extra_args
        assert robocopy_service._settings.delete_src
        assert robocopy_service._settings.overwrite
        assert not robocopy_service._settings.force_dir

    def test_transfer(self, mock_ui_helper, robocopy_service):
        with patch("src.clabe.data_transfer.robocopy.subprocess.Popen") as mock_popen:
            mock_ui_helper._prompt_yes_no_question.return_value = True
            mock_process = MagicMock()
            mock_process.wait.return_value = 0
            mock_popen.return_value = mock_process
            robocopy_service.transfer()

    def test_solve_src_dst_mapping_single_path(self, robocopy_service, robocopy_source, robocopy_settings):
        result = robocopy_service._solve_src_dst_mapping(robocopy_source, robocopy_settings.destination)
        assert result == {Path(robocopy_source): Path(robocopy_settings.destination)}

    def test_solve_src_dst_mapping_dict(self, robocopy_service, robocopy_source, robocopy_settings):
        source_dict = {robocopy_source: robocopy_settings.destination}
        result = robocopy_service._solve_src_dst_mapping(source_dict, None)
        assert result == source_dict

    def test_solve_src_dst_mapping_invalid(self, robocopy_service, robocopy_source):
        with pytest.raises(ValueError):
            robocopy_service._solve_src_dst_mapping(robocopy_source, None)



import logging
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

from clabe.logging_helper import add_file_handler, aibs


@pytest.fixture
def logger():
    test_logger = logging.getLogger("test_logger")
    test_logger.handlers = []  # Clear existing handlers
    return test_logger


@pytest.fixture
def settings():
    return aibs.AibsLogServerHandlerSettings(
        project_name="test_project",
        version="0.1.0",
        host="localhost",
        port=12345,
        rig_id="test_rig",
        comp_id="test_comp",
    )


class TestLoggingHelper:
    @patch("logging.FileHandler")
    def test_default_logger_builder_with_output_path(self, mock_file_handler, logger):
        mock_file_handler_instance = MagicMock()
        mock_file_handler.return_value = mock_file_handler_instance

        output_path = Path("/tmp/fake/path/to/logfile.log")
        result_logger = add_file_handler(logger, output_path)

        assert len(result_logger.handlers) == 1
        assert result_logger.handlers[0] == mock_file_handler_instance
        mock_file_handler.assert_called_once_with(output_path, encoding="utf-8", mode="w")

    @patch("clabe.logging_helper.aibs.AibsLogServerHandler")
    def test_add_log_server_handler(self, mock_log_server_handler, logger, settings):
        mock_log_server_handler_instance = MagicMock()
        mock_log_server_handler.return_value = mock_log_server_handler_instance

        result_logger = aibs.add_handler(logger, settings)

        assert len(result_logger.handlers) == 1
        assert result_logger.handlers[0] == mock_log_server_handler_instance
        mock_log_server_handler.assert_called_once_with(settings=settings)



from __future__ import annotations

import abc
import logging
from typing import Any, Generic, Optional, TypeVar

from ..services import Service

logger = logging.getLogger(__name__)


T = TypeVar("T")

TMapTo = TypeVar("TMapTo", bound=Any)


class DataMapper(Service, abc.ABC, Generic[TMapTo]):
    """
    Abstract base class for data mappers.

    This class defines the interface for mapping data from various sources to specific
    target formats or schemas. It provides a generic framework for data transformation
    and validation operations.

    Attributes:
        _mapped (Optional[TMapTo]): The mapped data object, set after successful mapping

    Example:
        ```python
        # Creating a custom data mapper
        class MyDataMapper(DataMapper[MyTargetType]):
            def map(self) -> MyTargetType:
                # Implementation specific mapping logic
                self._mapped = MyTargetType(...)
                return self._mapped

            def is_mapped(self) -> bool:
                return self._mapped is not None

            @property
            def mapped(self) -> MyTargetType:
                if not self.is_mapped():
                    raise ValueError("Data not yet mapped")
                return self._mapped

        # Using the mapper
        mapper = MyDataMapper()
        result = mapper.map()
        if mapper.is_mapped():
            data = mapper.mapped
        ```
    """

    _mapped: Optional[TMapTo]

    @abc.abstractmethod
    def map(self) -> TMapTo:
        """
        Maps data to the target schema or format.

        This method should contain the core logic for transforming input data
        into the target format specified by the TMapTo type parameter.

        Returns:
            TMapTo: The mapped data object
        """
        pass

    def is_mapped(self) -> bool:
        """
        Checks if the data has been successfully mapped.

        This method should verify whether the mapping operation has been completed
        and the data is available in the target format.

        Returns:
            bool: True if the data is mapped, False otherwise
        """
        return self._mapped is not None

    @property
    def mapped(self) -> TMapTo:
        """
        Retrieves the mapped data object.

        This property should return the successfully mapped data object.
        Implementations should ensure that mapping has been completed before
        returning the data.

        Returns:
            TMapTo: The mapped data object

        Raises:
            ValueError: If the data has not been mapped yet.
        """
        if not self.is_mapped():
            raise ValueError("Data not yet mapped")
        assert self._mapped is not None, "Mapped data should not be None"
        return self._mapped
